# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np  # noqa: F401
import pytest  # noqa: F401

import awkward as ak


def test():
    array = ak.Array([None, [1, 2, 3, [None, {"x": [None, 2], "y": [1, 4]}]]])

    assert ak.fill_none(array, -1.0, axis=None).to_list() == [
        -1.0,
        [1, 2, 3, [-1.0, {"x": [-1.0, 2], "y": [1, 4]}]],
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
__all__ = ("from_avro_file",)

from os import PathLike, fsdecode

import awkward as ak
from awkward._dispatch import high_level_function
from awkward._nplikes.numpylike import NumpyMetadata

np = NumpyMetadata.instance()


@high_level_function
def from_avro_file(
    file, limit_entries=None, *, debug_forth=False, highlevel=True, behavior=None
):
    """
    Args:
        file (path-like or file-like object): Avro file to be read as Awkward Array.
        limit_entries (int): The number of rows of the Avro file to be read into the Awkward Array.
        debug_forth (bool): If True, prints the generated Forth code for debugging.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Reads Avro files as Awkward Arrays.

    Internally this function uses AwkwardForth DSL. The function recursively parses the Avro schema, generates
    Awkward form and Forth code for that specific Avro file and then reads it.
    """
    import awkward._connect.avro

    if isinstance(file, (str, bytes, PathLike)):
        file = fsdecode(file)
        with open(file, "rb") as opened_file:
            form, length, container = awkward._connect.avro.ReadAvroFT(
                opened_file, limit_entries, debug_forth
            ).outcontents
            return _impl(form, length, container, highlevel, behavior)

    else:
        if not hasattr(file, "read") or not hasattr(file, "seek"):
            raise TypeError(
                "'file' must either be a filename string or be a file-like object with 'read' and 'seek' methods"
            )
        else:
            form, length, container = awkward._connect.avro.ReadAvroFT(
                file, limit_entries, debug_forth
            ).outarr
            return _impl(form, length, container, highlevel, behavior)


def _impl(form, length, container, highlevel, behavior):
    return ak.operations.ak_from_buffers._impl(
        form=form,
        length=length,
        container=container,
        buffer_key="{form_key}-{attribute}",
        backend="cpu",
        byteorder=ak._util.native_byteorder,
        highlevel=highlevel,
        behavior=behavior,
        simplify=True,
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
__all__ = ("categories",)
import awkward as ak
from awkward._behavior import behavior_of
from awkward._dispatch import high_level_function
from awkward._layout import wrap_layout


@high_level_function
def categories(array, highlevel=True):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.

    If the `array` is categorical (contains #ak.contents.IndexedArray or
    #ak.contents.IndexedOptionArray labeled with parameter
    `"__array__" = "categorical"`), then this function returns its categories.

    See also #ak.is_categorical, #ak.to_categorical, #ak.from_categorical.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, highlevel)


def _impl(array, highlevel):
    output = [None]

    def action(layout, **kwargs):
        if layout.parameter("__array__") == "categorical":
            output[0] = layout.content
            return layout

        else:
            return None

    layout = ak.operations.to_layout(array, allow_record=False, allow_other=False)
    behavior = behavior_of(array)
    ak._do.recursively_apply(layout, action, behavior=behavior)

    return wrap_layout(output[0], behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test_reducer():
    # broadcast_tooffsets64 was changed, and stopped trimming the content
    # this test ensures that things still work!
    a = ak.Array(
        ak.contents.ListOffsetArray(
            ak.index.Index64(np.array([1, 2, 3], dtype=np.int64)),
            ak.contents.NumpyArray(
                np.array([1, 1, 2, 1, 2, 3, 1, 2, 3, 4, 1, 2, 3, 4, 5], dtype=np.int64)
            ),
        )
    )
    # Perform broadcast, check result works!
    assert ak.all(a == [[1], [2]])



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import pytest  # noqa: F401

import awkward as ak


def test():
    left = ak.Array([0, 4, 5])
    right = ak.Array([[0, 1, 2, 3], [5, 6, 7, 2], [9, 8, 2, 0]])

    left_result, right_result = ak.broadcast_arrays(left, right, depth_limit=4)
    assert left_result.to_list() == [[0, 0, 0, 0], [4, 4, 4, 4], [5, 5, 5, 5]]
    assert right_result.to_list() == [[0, 1, 2, 3], [5, 6, 7, 2], [9, 8, 2, 0]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest

import awkward as ak


def test_wrap_index_cupy():
    cp = pytest.importorskip("cupy")
    data = cp.arange(10, dtype=cp.int64)
    index = ak.index.Index64(data)
    other_data = cp.asarray(index)
    assert cp.shares_memory(data, other_data)


def test_wrap_index_numpy():
    data = np.arange(10, dtype=np.int64)
    index = ak.index.Index64(data)
    other_data = np.asarray(index)
    assert np.shares_memory(data, other_data)


def test_wrap_bare_list():
    data = [1, 2, 3, 4, 5]
    index = ak.index.Index64(data)
    other_data = np.asarray(index)
    assert other_data.tolist() == data



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
from collections.abc import Iterable

import awkward as ak
from awkward._parameters import type_parameters_equal
from awkward._regularize import is_integer
from awkward._typing import JSONSerializable, final
from awkward._util import UNSET
from awkward.forms.form import Form


@final
class RecordForm(Form):
    is_record = True

    def __init__(
        self,
        contents,
        fields,
        *,
        parameters=None,
        form_key=None,
    ):
        if not isinstance(contents, Iterable):
            raise TypeError(
                "{} 'contents' must be iterable, not {}".format(
                    type(self).__name__, repr(contents)
                )
            )
        for content in contents:
            if not isinstance(content, Form):
                raise TypeError(
                    "{} all 'contents' must be Form subclasses, not {}".format(
                        type(self).__name__, repr(content)
                    )
                )
        if fields is not None and not isinstance(fields, Iterable):
            raise TypeError(
                "{} 'fields' must be iterable, not {}".format(
                    type(self).__name__, repr(contents)
                )
            )

        self._fields = None if fields is None else list(fields)
        self._contents = list(contents)
        self._init(parameters=parameters, form_key=form_key)

    @property
    def contents(self):
        return self._contents

    @property
    def fields(self):
        if self._fields is None:
            return [str(i) for i in range(len(self._contents))]
        else:
            return self._fields

    def copy(
        self,
        contents=UNSET,
        fields=UNSET,
        *,
        parameters=UNSET,
        form_key=UNSET,
    ):
        return RecordForm(
            self._contents if contents is UNSET else contents,
            self._fields if fields is UNSET else fields,
            parameters=self._parameters if parameters is UNSET else parameters,
            form_key=self._form_key if form_key is UNSET else form_key,
        )

    @classmethod
    def simplified(
        cls,
        contents,
        fields,
        *,
        parameters=None,
        form_key=None,
    ):
        return cls(contents, fields, parameters=parameters, form_key=form_key)

    @property
    def is_tuple(self):
        return self._fields is None

    def __repr__(self):
        args = [repr(self._contents), repr(self._fields), *self._repr_args()]
        return "{}({})".format(type(self).__name__, ", ".join(args))

    def index_to_field(self, index):
        if 0 <= index < len(self._contents):
            if self._fields is None:
                return str(index)
            else:
                return self._fields[index]
        else:
            raise IndexError(
                "no index {} in record with {} fields".format(
                    index, len(self._contents)
                )
            )

    def field_to_index(self, field):
        if self._fields is None:
            try:
                i = int(field)
            except ValueError:
                pass
            else:
                if 0 <= i < len(self._contents):
                    return i
        else:
            try:
                i = self._fields.index(field)
            except ValueError:
                pass
            else:
                return i
        raise ak._errors.FieldNotFoundError(
            "no field {} in record with {} fields".format(
                repr(field), len(self._contents)
            )
        )

    def has_field(self, field):
        if self._fields is None:
            try:
                i = int(field)
            except ValueError:
                return False
            else:
                return 0 <= i < len(self._contents)
        else:
            return field in self._fields

    def content(self, index_or_field):
        if is_integer(index_or_field):
            index = index_or_field
        elif isinstance(index_or_field, str):
            index = self.field_to_index(index_or_field)
        else:
            raise TypeError(
                "index_or_field must be an integer (index) or string (field), not {}".format(
                    repr(index_or_field)
                )
            )
        return self._contents[index]

    def _to_dict_part(self, verbose, toplevel):
        out = {"class": "RecordArray"}

        contents_tolist = [
            content._to_dict_part(verbose, toplevel=False) for content in self._contents
        ]
        if self._fields is not None:
            out["fields"] = list(self._fields)
        else:
            out["fields"] = None

        out["contents"] = contents_tolist
        return self._to_dict_extra(out, verbose)

    @property
    def type(self):
        return ak.types.RecordType(
            [x.type for x in self._contents],
            self._fields,
            parameters=self._parameters,
        )

    def __eq__(self, other):
        if isinstance(other, RecordForm):
            if (
                self._form_key == other._form_key
                and self.is_tuple == other.is_tuple
                and len(self._contents) == len(other._contents)
                and type_parameters_equal(self._parameters, other._parameters)
            ):
                if self.is_tuple:
                    return self._contents == other._contents
                else:
                    return dict(zip(self._fields, self._contents)) == dict(
                        zip(other._fields, other._contents)
                    )
            else:
                return False
        else:
            return False

    def purelist_parameters(self, *keys: str) -> JSONSerializable:
        if self._parameters is not None:
            for key in keys:
                if key in self._parameters:
                    return self._parameters[key]

    @property
    def purelist_isregular(self):
        return True

    @property
    def purelist_depth(self):
        return 1

    @property
    def is_identity_like(self):
        return False

    @property
    def minmax_depth(self):
        if len(self._contents) == 0:
            return (1, 1)
        mins, maxs = [], []
        for content in self._contents:
            mindepth, maxdepth = content.minmax_depth
            mins.append(mindepth)
            maxs.append(maxdepth)
        return (min(mins), max(maxs))

    @property
    def branch_depth(self):
        if len(self._contents) == 0:
            return (False, 1)
        anybranch = False
        mindepth = None
        for content in self._contents:
            branch, depth = content.branch_depth
            if mindepth is None:
                mindepth = depth
            if branch or mindepth != depth:
                anybranch = True
            if mindepth > depth:
                mindepth = depth
        return (anybranch, mindepth)

    @property
    def dimension_optiontype(self):
        return False

    def _columns(self, path, output, list_indicator):
        for content, field in zip(self._contents, self.fields):
            content._columns((*path, field), output, list_indicator)

    def _prune_columns(self, is_inside_record_or_union: bool):
        contents = []
        fields = []
        for content, field in zip(self._contents, self.fields):
            next_content = content._prune_columns(True)
            if next_content is None:
                continue
            contents.append(next_content)
            fields.append(field)

        if fields or not is_inside_record_or_union:
            return RecordForm(
                contents,
                fields,
                parameters=self._parameters,
                form_key=self._form_key,
            )
        # If all subtrees are pruned (or we have no subtrees!), then we should prune this form too
        else:
            return None

    def _select_columns(self, match_specifier):
        contents = []
        fields = []
        for content, field in zip(self._contents, self.fields):
            # Try and match this field, allowing derived matcher to match any field if empty
            next_match_specifier = match_specifier(field, next_match_if_empty=True)
            if next_match_specifier is None:
                continue

            # NOTE: we could optimise this by avoiding column selection if we know the entire subtree will match
            #       this is given by `next_match_specifier.is_empty`, *but* we also want to perform column pruning
            #       meaning this optimisation is less useful, we we'd require a special prune-only pass
            next_content = content._select_columns(next_match_specifier)
            contents.append(next_content)
            fields.append(field)

        return RecordForm(
            contents,
            fields,
            parameters=self._parameters,
            form_key=self._form_key,
        )

    def _column_types(self):
        return sum((x._column_types() for x in self._contents), ())



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak

to_list = ak.operations.to_list


def test():
    array = ak.contents.RegularArray(
        ak.contents.NumpyArray(np.empty(0, dtype=np.int32)),
        size=0,
        zeros_length=1,
    )
    packed = ak.operations.to_packed(array)
    assert to_list(packed) == [[]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest

import awkward as ak

jax = pytest.importorskip("jax")
jax.config.update("jax_platform_name", "cpu")
jax.config.update("jax_enable_x64", True)

ak.jax.register_and_check()

# #### ak.contents.NumpyArray ####

test_numpyarray = ak.Array(np.arange(10, dtype=np.float64), backend="jax")
test_numpyarray_tangent = ak.Array(np.arange(10, dtype=np.float64), backend="jax")

test_numpyarray_jax = jax.numpy.arange(10, dtype=np.float64)
test_numpyarray_tangent_jax = jax.numpy.arange(10, dtype=np.float64)


def test_numpyarray_grad_1():
    def func_numpyarray_1(x):
        return x[4] ** 2

    value_jvp, jvp_grad = jax.jvp(
        func_numpyarray_1, (test_numpyarray,), (test_numpyarray_tangent,)
    )
    value_jvp_jax, jvp_grad_jax = jax.jvp(
        func_numpyarray_1, (test_numpyarray_jax,), (test_numpyarray_tangent_jax,)
    )

    value_vjp, vjp_func = jax.vjp(func_numpyarray_1, test_numpyarray)
    value_vjp_jax, vjp_func_jax = jax.vjp(func_numpyarray_1, test_numpyarray_jax)

    assert value_jvp == value_jvp_jax
    assert value_vjp == value_vjp_jax
    assert jvp_grad == jvp_grad_jax
    assert (
        ak.to_list(vjp_func(value_vjp)[0]) == (vjp_func_jax(value_vjp_jax)[0]).tolist()
    )


def test_numpyarray_grad_2():
    def func_numpyarray_2(x):
        return x[2:5] ** 2 + x[1:4] ** 2

    value_jvp, jvp_grad = jax.jvp(
        func_numpyarray_2, (test_numpyarray,), (test_numpyarray_tangent,)
    )
    value_jvp_jax, jvp_grad_jax = jax.jvp(
        func_numpyarray_2, (test_numpyarray_jax,), (test_numpyarray_tangent_jax,)
    )

    value_vjp, vjp_func = jax.vjp(func_numpyarray_2, test_numpyarray)
    value_vjp_jax, vjp_func_jax = jax.vjp(func_numpyarray_2, test_numpyarray_jax)

    assert ak.to_list(value_jvp) == value_jvp_jax.tolist()
    assert ak.to_list(value_vjp) == value_vjp_jax.tolist()
    assert ak.to_list(jvp_grad) == jvp_grad_jax.tolist()
    assert (
        ak.to_list(vjp_func(value_vjp)[0]) == (vjp_func_jax(value_vjp_jax)[0]).tolist()
    )


def test_numpyarray_grad_3():
    def func_numpyarray_3(x):
        return x[::-1]

    value_jvp, jvp_grad = jax.jvp(
        func_numpyarray_3, (test_numpyarray,), (test_numpyarray_tangent,)
    )
    value_jvp_jax, jvp_grad_jax = jax.jvp(
        func_numpyarray_3, (test_numpyarray_jax,), (test_numpyarray_tangent_jax,)
    )

    value_vjp, vjp_func = jax.vjp(func_numpyarray_3, test_numpyarray)
    value_vjp_jax, vjp_func_jax = jax.vjp(func_numpyarray_3, test_numpyarray_jax)

    assert ak.to_list(value_jvp) == value_jvp_jax.tolist()
    assert ak.to_list(value_vjp) == value_vjp_jax.tolist()
    assert ak.to_list(jvp_grad) == jvp_grad_jax.tolist()
    assert (
        ak.to_list(vjp_func(value_vjp)[0]) == (vjp_func_jax(value_vjp_jax)[0]).tolist()
    )


def test_numpyarray_grad_4():
    def func_numpyarray_4(x):
        return x[2:5] ** 2 * x[1:4] ** 2

    value_jvp, jvp_grad = jax.jvp(
        func_numpyarray_4, (test_numpyarray,), (test_numpyarray_tangent,)
    )
    value_jvp_jax, jvp_grad_jax = jax.jvp(
        func_numpyarray_4, (test_numpyarray_jax,), (test_numpyarray_tangent_jax,)
    )

    value_vjp, vjp_func = jax.vjp(func_numpyarray_4, test_numpyarray)
    value_vjp_jax, vjp_func_jax = jax.vjp(func_numpyarray_4, test_numpyarray_jax)

    assert ak.to_list(value_jvp) == value_jvp_jax.tolist()
    assert ak.to_list(value_vjp) == value_vjp_jax.tolist()
    assert ak.to_list(jvp_grad) == jvp_grad_jax.tolist()
    assert (
        ak.to_list(vjp_func(value_vjp)[0]) == (vjp_func_jax(value_vjp_jax)[0]).tolist()
    )


test_regulararray = ak.Array(
    [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]], backend="jax"
)
test_regulararray_tangent = ak.Array(
    [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]], backend="jax"
)

test_regulararray_jax = jax.numpy.array(
    [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]], dtype=np.float64
)
test_regulararray_tangent_jax = jax.numpy.array(
    [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]], dtype=np.float64
)


def test_regular_array_1():
    def func_regulararray_1(x):
        return x[2] * 2

    value_jvp, jvp_grad = jax.jvp(
        func_regulararray_1, (test_regulararray,), (test_regulararray_tangent,)
    )
    value_jvp_jax, jvp_grad_jax = jax.jvp(
        func_regulararray_1, (test_regulararray_jax,), (test_regulararray_tangent_jax,)
    )

    value_vjp, vjp_func = jax.vjp(func_regulararray_1, test_regulararray)
    value_vjp_jax, vjp_func_jax = jax.vjp(func_regulararray_1, test_regulararray_jax)

    assert ak.to_list(value_jvp) == value_jvp_jax.tolist()
    assert ak.to_list(value_vjp) == value_vjp_jax.tolist()
    assert ak.to_list(jvp_grad) == jvp_grad_jax.tolist()
    assert (
        ak.to_list(vjp_func(value_vjp)[0]) == (vjp_func_jax(value_vjp_jax)[0]).tolist()
    )


def test_regular_array_2():
    def func_regulararray_2(x):
        return x * x

    value_jvp, jvp_grad = jax.jvp(
        func_regulararray_2, (test_regulararray,), (test_regulararray_tangent,)
    )
    value_jvp_jax, jvp_grad_jax = jax.jvp(
        func_regulararray_2, (test_regulararray_jax,), (test_regulararray_tangent_jax,)
    )

    value_vjp, vjp_func = jax.vjp(func_regulararray_2, test_regulararray)
    value_vjp_jax, vjp_func_jax = jax.vjp(func_regulararray_2, test_regulararray_jax)

    assert ak.to_list(value_jvp) == value_jvp_jax.tolist()
    assert ak.to_list(value_vjp) == value_vjp_jax.tolist()
    assert ak.to_list(jvp_grad) == jvp_grad_jax.tolist()
    assert (
        ak.to_list(vjp_func(value_vjp)[0]) == (vjp_func_jax(value_vjp_jax)[0]).tolist()
    )


def test_regular_array_3():
    def func_regular_array_3(x):
        return x[0, 0] * x[2, 1]

    value_jvp, jvp_grad = jax.jvp(
        func_regular_array_3, (test_regulararray,), (test_regulararray_tangent,)
    )
    value_jvp_jax, jvp_grad_jax = jax.jvp(
        func_regular_array_3, (test_regulararray_jax,), (test_regulararray_tangent_jax,)
    )

    value_vjp, vjp_func = jax.vjp(func_regular_array_3, test_regulararray)
    value_vjp_jax, vjp_func_jax = jax.vjp(func_regular_array_3, test_regulararray_jax)

    assert ak.to_list(value_jvp) == value_jvp_jax.tolist()
    assert ak.to_list(value_vjp) == value_vjp_jax.tolist()
    assert ak.to_list(jvp_grad) == jvp_grad_jax.tolist()
    assert (
        ak.to_list(vjp_func(value_vjp)[0]) == (vjp_func_jax(value_vjp_jax)[0]).tolist()
    )


def test_regular_array_4():
    def func_regular_array_4(x):
        return x[::-1] ** 2

    value_jvp, jvp_grad = jax.jvp(
        func_regular_array_4, (test_regulararray,), (test_regulararray_tangent,)
    )
    value_jvp_jax, jvp_grad_jax = jax.jvp(
        func_regular_array_4, (test_regulararray_jax,), (test_regulararray_tangent_jax,)
    )

    value_vjp, vjp_func = jax.vjp(func_regular_array_4, test_regulararray)
    value_vjp_jax, vjp_func_jax = jax.vjp(func_regular_array_4, test_regulararray_jax)

    assert ak.to_list(value_jvp) == value_jvp_jax.tolist()
    assert ak.to_list(value_vjp) == value_vjp_jax.tolist()
    assert ak.to_list(jvp_grad) == jvp_grad_jax.tolist()
    assert (
        ak.to_list(vjp_func(value_vjp)[0]) == (vjp_func_jax(value_vjp_jax)[0]).tolist()
    )


def test_regular_array_5():
    def func_regular_array_5(x):
        return 2 * x[:-1]

    value_jvp, jvp_grad = jax.jvp(
        func_regular_array_5, (test_regulararray,), (test_regulararray_tangent,)
    )
    value_jvp_jax, jvp_grad_jax = jax.jvp(
        func_regular_array_5, (test_regulararray_jax,), (test_regulararray_tangent_jax,)
    )

    value_vjp, vjp_func = jax.vjp(func_regular_array_5, test_regulararray)
    value_vjp_jax, vjp_func_jax = jax.vjp(func_regular_array_5, test_regulararray_jax)

    assert ak.to_list(value_jvp) == value_jvp_jax.tolist()
    assert ak.to_list(value_vjp) == value_vjp_jax.tolist()
    assert ak.to_list(jvp_grad) == jvp_grad_jax.tolist()
    assert (
        ak.to_list(vjp_func(value_vjp)[0]) == (vjp_func_jax(value_vjp_jax)[0]).tolist()
    )


def test_regular_array_6():
    def func_regular_array_6(x):
        return x[0][0] * x[2][1]

    value_jvp, jvp_grad = jax.jvp(
        func_regular_array_6, (test_regulararray,), (test_regulararray_tangent,)
    )
    value_jvp_jax, jvp_grad_jax = jax.jvp(
        func_regular_array_6, (test_regulararray_jax,), (test_regulararray_tangent_jax,)
    )

    value_vjp, vjp_func = jax.vjp(func_regular_array_6, test_regulararray)
    value_vjp_jax, vjp_func_jax = jax.vjp(func_regular_array_6, test_regulararray_jax)

    assert ak.to_list(value_jvp) == value_jvp_jax.tolist()
    assert ak.to_list(value_vjp) == value_vjp_jax.tolist()
    assert ak.to_list(jvp_grad) == jvp_grad_jax.tolist()
    assert (
        ak.to_list(vjp_func(value_vjp)[0]) == (vjp_func_jax(value_vjp_jax)[0]).tolist()
    )


test_recordarray = ak.Array(
    [
        [{"x": 1.1, "y": [1.0]}, {"x": 2.2, "y": [1.0, 2.2]}],
        [],
        [{"x": 3.3, "y": [1.0, 2.0, 3.0]}],
    ],
    backend="jax",
)
test_recordarray_tangent = ak.Array(
    [
        [{"x": 0.0, "y": [1.0]}, {"x": 2.0, "y": [1.5, 0.0]}],
        [],
        [{"x": 1.5, "y": [2.0, 0.5, 1.0]}],
    ],
    backend="jax",
)


def test_recordarray_1():
    def func_recordarray_1(x):
        return 2 * x.y[2][0][1] + 10

    value_jvp, jvp_grad = jax.jvp(
        func_recordarray_1, (test_recordarray,), (test_recordarray_tangent,)
    )
    value_vjp, vjp_func = jax.vjp(func_recordarray_1, test_recordarray)
    assert ak.to_list(value_jvp) == 14.0
    assert ak.to_list(value_vjp) == 14.0
    assert ak.to_list(jvp_grad) == 1.0
    assert ak.to_list(vjp_func(value_vjp)[0]) == [
        [{"x": 0.0, "y": [0.0]}, {"x": 0.0, "y": [0.0, 0.0]}],
        [],
        [{"x": 0.0, "y": [0.0, 28.0, 0.0]}],
    ]


def test_recordarray_2():
    def func_recordarray_2(x):
        return 2 * x.y[2][0] + 10

    value_jvp, jvp_grad = jax.jvp(
        func_recordarray_2, (test_recordarray,), (test_recordarray_tangent,)
    )
    value_vjp, vjp_func = jax.vjp(func_recordarray_2, test_recordarray)
    assert ak.to_list(value_jvp) == [12.0, 14.0, 16.0]
    assert ak.to_list(value_vjp) == [12.0, 14.0, 16.0]
    assert ak.to_list(jvp_grad) == [4.0, 1.0, 2.0]
    assert ak.to_list(vjp_func(value_vjp)[0]) == [
        [{"x": 0.0, "y": [0.0]}, {"x": 0.0, "y": [0.0, 0.0]}],
        [],
        [{"x": 0.0, "y": [24.0, 28.0, 32.0]}],
    ]


def test_recordarray_3():
    def test_recordarray_3(x):
        return 2 * x.y[0][0] ** 2

    value_jvp, jvp_grad = jax.jvp(
        test_recordarray_3, (test_recordarray,), (test_recordarray_tangent,)
    )
    value_vjp, vjp_func = jax.vjp(test_recordarray_3, test_recordarray)
    assert ak.to_list(value_jvp) == [2.0]
    assert ak.to_list(value_vjp) == [2.0]
    assert ak.to_list(jvp_grad) == [4.0]
    assert ak.to_list(vjp_func(value_vjp)[0]) == [
        [{"x": 0.0, "y": [8.0]}, {"x": 0.0, "y": [0.0, 0.0]}],
        [],
        [{"x": 0.0, "y": [0.0, 0.0, 0.0]}],
    ]


def test_recordarray_4():
    def test_recordarray_4(x):
        return 2 * x.y[2] + 10

    value_jvp, jvp_grad = jax.jvp(
        test_recordarray_4, (test_recordarray,), (test_recordarray_tangent,)
    )
    value_vjp, vjp_func = jax.vjp(test_recordarray_4, test_recordarray)
    assert ak.to_list(value_jvp) == [[12.0, 14.0, 16.0]]
    assert ak.to_list(value_vjp) == [[12.0, 14.0, 16.0]]
    assert ak.to_list(jvp_grad) == [[4.0, 1.0, 2.0]]
    assert ak.to_list(vjp_func(value_vjp)[0]) == [
        [{"x": 0.0, "y": [0.0]}, {"x": 0.0, "y": [0.0, 0.0]}],
        [],
        [{"x": 0.0, "y": [24.0, 28.0, 32.0]}],
    ]


def test_recordarray_5():
    def test_recordarray_5(x):
        return 2 * x.y

    value_jvp, jvp_grad = jax.jvp(
        test_recordarray_5, (test_recordarray,), (test_recordarray_tangent,)
    )
    value_vjp, vjp_func = jax.vjp(test_recordarray_5, test_recordarray)
    assert ak.to_list(value_jvp) == [[[2.0], [2.0, 4.4]], [], [[2.0, 4.0, 6.0]]]
    assert ak.to_list(value_vjp) == [[[2.0], [2.0, 4.4]], [], [[2.0, 4.0, 6.0]]]
    assert ak.to_list(jvp_grad) == [[[2.0], [3.0, 0.0]], [], [[4.0, 1.0, 2.0]]]
    assert ak.to_list(vjp_func(value_vjp)[0]) == [
        [{"x": 0.0, "y": [4.0]}, {"x": 0.0, "y": [4.0, 8.8]}],
        [],
        [{"x": 0.0, "y": [4.0, 8.0, 12.0]}],
    ]


def test_recordarray_6():
    def test_recordarray_6(x):
        return 2 * x.y**2

    value_jvp, jvp_grad = jax.jvp(
        test_recordarray_6, (test_recordarray,), (test_recordarray_tangent,)
    )
    value_vjp, vjp_func = jax.vjp(test_recordarray_6, test_recordarray)
    assert ak.to_list(value_jvp) == [
        [[2.0], [2.0, 9.680000000000001]],
        [],
        [[2.0, 8.0, 18.0]],
    ]
    assert ak.to_list(value_vjp) == [
        [[2.0], [2.0, 9.680000000000001]],
        [],
        [[2.0, 8.0, 18.0]],
    ]
    assert ak.to_list(jvp_grad) == [[[4.0], [6.0, 0.0]], [], [[8.0, 4.0, 12.0]]]
    assert ak.to_list(vjp_func(value_vjp)[0]) == [
        [{"x": 0.0, "y": [8.0]}, {"x": 0.0, "y": [8.0, 85.18400000000003]}],
        [],
        [{"x": 0.0, "y": [8.0, 64.0, 216.0]}],
    ]


def test_recordarray_7():
    def test_recordarray_7(x):
        return 2 * x.y[2, 0, 1] + 10

    value_jvp, jvp_grad = jax.jvp(
        test_recordarray_7, (test_recordarray,), (test_recordarray_tangent,)
    )
    value_vjp, vjp_func = jax.vjp(test_recordarray_7, test_recordarray)
    assert ak.to_list(value_jvp) == 14.0
    assert ak.to_list(value_vjp) == 14.0
    assert ak.to_list(jvp_grad) == 1.0
    assert ak.to_list(vjp_func(value_vjp)[0]) == [
        [{"x": 0.0, "y": [0.0]}, {"x": 0.0, "y": [0.0, 0.0]}],
        [],
        [{"x": 0.0, "y": [0.0, 28.0, 0.0]}],
    ]


def test_recordarray_8():
    def func_recordarray_8(x):
        return 2 * x.y[2, 0] + 10

    value_jvp, jvp_grad = jax.jvp(
        func_recordarray_8, (test_recordarray,), (test_recordarray_tangent,)
    )
    value_vjp, vjp_func = jax.vjp(func_recordarray_8, test_recordarray)
    assert ak.to_list(value_jvp) == [12.0, 14.0, 16.0]
    assert ak.to_list(value_vjp) == [12.0, 14.0, 16.0]
    assert ak.to_list(jvp_grad) == [4.0, 1.0, 2.0]
    assert ak.to_list(vjp_func(value_vjp)[0]) == [
        [{"x": 0.0, "y": [0.0]}, {"x": 0.0, "y": [0.0, 0.0]}],
        [],
        [{"x": 0.0, "y": [24.0, 28.0, 32.0]}],
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
from __future__ import annotations

import awkward as ak
from awkward._nplikes.array_module import ArrayModuleNumpyLike
from awkward._nplikes.dispatch import register_nplike
from awkward._nplikes.numpylike import ArrayLike
from awkward._nplikes.shape import ShapeItem
from awkward._typing import Final


@register_nplike
class Jax(ArrayModuleNumpyLike):
    is_eager: Final = True

    def __init__(self):
        jax = ak.jax.import_jax()
        self._module = jax.numpy

    @property
    def ma(self):
        raise ValueError(
            "JAX arrays cannot have missing values until JAX implements "
            "numpy.ma.MaskedArray"
        )

    @property
    def char(self):
        raise ValueError(
            "JAX arrays cannot do string manipulations until JAX implements "
            "numpy.char"
        )

    @property
    def ndarray(self):
        return self._module.ndarray

    @classmethod
    def is_own_array_type(cls, type_: type) -> bool:
        """
        Args:
            type_: object to test

        Return `True` if the given object is a jax buffer, otherwise `False`.

        """
        return cls.is_array_type(type_) or cls.is_tracer_type(type_)

    @classmethod
    def is_array_type(cls, type_: type) -> bool:
        """
        Args:
            type_: object to test

        Return `True` if the given object is a jax buffer, otherwise `False`.

        """
        module, _, suffix = type_.__module__.partition(".")
        return module == "jaxlib"

    @classmethod
    def is_tracer_type(cls, type_: type) -> bool:
        """
        Args:
            type_: object to test

        Return `True` if the given object is a jax tracer, otherwise `False`.

        """
        module, _, suffix = type_.__module__.partition(".")
        return module == "jax"

    def is_c_contiguous(self, x: ArrayLike) -> bool:
        return True

    def ascontiguousarray(self, x: ArrayLike) -> ArrayLike:
        return x

    def strides(self, x: ArrayLike) -> tuple[ShapeItem, ...]:
        x.touch_shape()
        out = (x._dtype.itemsize,)
        for item in reversed(x._shape):
            out = (item * out[0], *out)
        return out



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
__all__ = ("pad_none",)
import awkward as ak
from awkward._dispatch import high_level_function
from awkward._layout import wrap_layout
from awkward._nplikes.numpylike import NumpyMetadata
from awkward._regularize import regularize_axis

np = NumpyMetadata.instance()


@high_level_function
def pad_none(array, target, axis=1, *, clip=False, highlevel=True, behavior=None):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        target (int): The intended length of the lists. If `clip=True`,
            the output lists will have exactly this length; otherwise,
            they will have *at least* this length.
        axis (int): The dimension at which this operation is applied. The
            outermost dimension is `0`, followed by `1`, etc., and negative
            values count backward from the innermost: `-1` is the innermost
            dimension, `-2` is the next level up, etc.
        clip (bool): If True, the output lists will have regular lengths
            (#ak.types.RegularType) of exactly `target`; otherwise the
            output lists will have in-principle variable lengths
            (#ak.types.ListType) of at least `target`.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Increase the lengths of lists to a target length by adding None values.

    Consider the following

        >>> array = ak.Array([[[1.1, 2.2, 3.3],
        ...                    [],
        ...                    [4.4, 5.5],
        ...                    [6.6]],
        ...                   [],
        ...                   [[7.7],
        ...                    [8.8, 9.9]
        ...                   ]])

    At `axis=0`, this operation pads the whole array, adding None at the
    outermost level:

        >>> ak.pad_none(array, 5, axis=0).show()
        [[[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6]],
         [],
         [[7.7], [8.8, 9.9]],
         None,
         None]

    At `axis=1`, this operation pads the first nested level:

        >>> ak.pad_none(array, 3, axis=1).show()
        [[[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6]],
         [None, None, None],
         [[7.7], [8.8, 9.9], None]]

    And so on for higher values of `axis`:

        >>> ak.pad_none(array, 2, axis=2).show()
        [[[1.1, 2.2, 3.3], [None, None], [4.4, 5.5], [6.6, None]],
         [],
         [[7.7, None], [8.8, 9.9]]]

    Note that the `clip` parameter not only determines whether the lengths are
    at least `target` or exactly `target`, it also determines the type of the
    output:

    * `clip=True` returns regular lists (#ak.types.RegularType), and
    * `clip=False` returns in-principle variable lengths
      (#ak.types.ListType).

    The in-principle variable-length lists might, in fact, all have the same
    length, but the type difference is significant, for instance in
    broadcasting rules (see #ak.broadcast_arrays).

    The difference between

        >>> ak.pad_none(array, 2, axis=2)
        <Array [[[1.1, 2.2, 3.3], ..., [...]], ...] type='3 * var * var * ?float64'>

    and

        >>> ak.pad_none(array, 2, axis=2, clip=True)
        <Array [[[1.1, 2.2], ..., [6.6, None]], ...] type='3 * var * 2 * ?float64'>

    is not just in the length of `[1.1, 2.2, 3.3]` vs `[1.1, 2.2]`, but also
    in the distinction between the following types.

        >>> ak.pad_none(array, 2, axis=2).type.show()
        3 * var * var * ?float64
        >>> ak.pad_none(array, 2, axis=2, clip=True).type.show()
        3 * var *   2 * ?float64
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, target, axis, clip, highlevel, behavior)


def _impl(array, target, axis, clip, highlevel, behavior):
    axis = regularize_axis(axis)
    layout = ak.operations.to_layout(array, allow_record=False, allow_other=False)
    out = ak._do.pad_none(layout, target, axis, clip=clip)

    return wrap_layout(out, behavior, highlevel, like=array)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np  # noqa: F401
import pytest

import awkward as ak

to_list = ak.operations.to_list


def test():
    a = ak.Array([{"this": 100}])
    b = ak.Array([{"this": 90, "that": 100}])
    c = ak.operations.concatenate((a, b))

    with pytest.raises(ValueError):
        ak.operations.unzip(c)

    a = ak.Array([{"this": 100}])
    b = ak.Array([{"this": 90}])
    c = ak.operations.concatenate((a, b))

    (tmp,) = ak.operations.unzip(c)

    assert tmp.to_list() == [100, 90]


def test_zip_and_unzip():
    x = ak.Array([[1, 2, 3], [], [4, 5], [6], [7, 8, 9, 10]])
    y = ak.Array([1.1, 2.2, 3.3, 4.4, 5.5])

    one = ak.operations.zip({"x": x, "y": y})
    two = ak.operations.zip({"x": x, "y": y}, depth_limit=1)
    xx, yy = ak.operations.unzip(two)
    assert isinstance(one.layout, ak.contents.Content)
    assert isinstance(two.layout, ak.contents.Content)
    assert isinstance(xx.layout, ak.contents.Content)
    assert isinstance(yy.layout, ak.contents.Content)
    assert to_list(one) == [
        [{"x": 1, "y": 1.1}, {"x": 2, "y": 1.1}, {"x": 3, "y": 1.1}],
        [],
        [{"x": 4, "y": 3.3}, {"x": 5, "y": 3.3}],
        [{"x": 6, "y": 4.4}],
        [
            {"x": 7, "y": 5.5},
            {"x": 8, "y": 5.5},
            {"x": 9, "y": 5.5},
            {"x": 10, "y": 5.5},
        ],
    ]
    assert to_list(two) == [
        {"x": [1, 2, 3], "y": 1.1},
        {"x": [], "y": 2.2},
        {"x": [4, 5], "y": 3.3},
        {"x": [6], "y": 4.4},
        {"x": [7, 8, 9, 10], "y": 5.5},
    ]
    assert to_list(xx) == [[1, 2, 3], [], [4, 5], [6], [7, 8, 9, 10]]
    assert to_list(yy) == [1.1, 2.2, 3.3, 4.4, 5.5]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np  # noqa: F401
import pytest  # noqa: F401

import awkward as ak


def test_argsort():
    array = ak.Array([[1.1, 2.2], [3.3, 3.1]])
    assert ak.operations.argsort(array).to_list() == [[0, 1], [1, 0]]
    assert str(ak.operations.type(ak.operations.argsort(array))) == "2 * var * int64"

    empty_array = ak.Array([[], []])
    assert ak.operations.argsort(empty_array).to_list() == [[], []]
    assert (
        str(ak.operations.type(ak.operations.argsort(empty_array))) == "2 * var * int64"
    )

    select_array = array[array > 5]
    assert select_array.to_list() == [[], []]
    assert str(ak.operations.type(select_array)) == "2 * var * float64"

    assert ak.operations.argsort(select_array).to_list() == [[], []]
    assert (
        str(ak.operations.type(ak.operations.argsort(select_array)))
        == "2 * var * int64"
    )

    assert ak.operations.argsort(array[array > 5]).to_list() == [[], []]
    assert (
        str(ak.operations.type(ak.operations.argsort(array[array > 5])))
        == "2 * var * int64"
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
__all__ = ("local_index",)
import awkward as ak
from awkward._dispatch import high_level_function
from awkward._layout import wrap_layout
from awkward._nplikes.numpylike import NumpyMetadata
from awkward._regularize import regularize_axis

np = NumpyMetadata.instance()


@high_level_function
def local_index(array, axis=-1, *, highlevel=True, behavior=None):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        axis (int): The dimension at which this operation is applied. The
            outermost dimension is `0`, followed by `1`, etc., and negative
            values count backward from the innermost: `-1` is the innermost
            dimension, `-2` is the next level up, etc.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    For example,

        >>> array = ak.Array([
        ...     [[0.0, 1.1, 2.2], []],
        ...     [[3.3, 4.4]],
        ...     [],
        ...     [[5.5], [], [6.6, 7.7, 8.8, 9.9]]])
        >>> ak.local_index(array, axis=0)
        <Array [0, 1, 2, 3] type='4 * int64'>
        >>> ak.local_index(array, axis=1)
        <Array [[0, 1], [0], [], [0, 1, 2]] type='4 * var * int64'>
        >>> ak.local_index(array, axis=2)
        <Array [[[0, 1, 2], []], ..., [[0], ..., [...]]] type='4 * var * var * int64'>

    Note that you can make a Pandas-style MultiIndex by calling this function on
    every axis.

        >>> multiindex = ak.zip([ak.local_index(array, i) for i in range(array.ndim)])
        >>> multiindex.show()
        [[[(0, 0, 0), (0, 0, 1), (0, 0, 2)], []],
         [[(1, 0, 0), (1, 0, 1)]],
         [],
         [[(3, 0, 0)], [], [(3, 2, 0), (3, 2, 1), (3, 2, 2), (3, 2, 3)]]]
        >>> ak.flatten(ak.flatten(multiindex)).show()
        [(0, 0, 0),
         (0, 0, 1),
         (0, 0, 2),
         (1, 0, 0),
         (1, 0, 1),
         (3, 0, 0),
         (3, 2, 0),
         (3, 2, 1),
         (3, 2, 2),
         (3, 2, 3)]

    But if you're interested in Pandas, you may want to use #ak.to_dataframe directly.

        >>> ak.to_dataframe(array)
                                    values
        entry subentry subsubentry
        0     0        0               0.0
                       1               1.1
                       2               2.2
        1     0        0               3.3
                       1               4.4
        3     0        0               5.5
              2        0               6.6
                       1               7.7
                       2               8.8
                       3               9.9
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, axis, highlevel, behavior)


def _impl(array, axis, highlevel, behavior):
    axis = regularize_axis(axis)
    layout = ak.operations.to_layout(array, allow_record=True, allow_other=False)
    out = ak._do.local_index(layout, axis)
    return wrap_layout(out, behavior, highlevel, like=array)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
__all__ = ("zip",)
from collections.abc import Mapping

import awkward as ak
from awkward._behavior import behavior_of
from awkward._dispatch import high_level_function
from awkward._layout import wrap_layout
from awkward._nplikes.numpylike import NumpyMetadata

np = NumpyMetadata.instance()


@high_level_function
def zip(
    arrays,
    depth_limit=None,
    *,
    parameters=None,
    with_name=None,
    right_broadcast=False,
    optiontype_outside_record=False,
    highlevel=True,
    behavior=None,
):
    """
    Args:
        arrays (dict or iterable of arrays): Each value in this dict or iterable
            can be any array-like data that #ak.to_layout recognizes.
        depth_limit (None or int): If None, attempt to fully broadcast the
            `array` to all levels. If an int, limit the number of dimensions
            that get broadcasted. The minimum value is `1`, for no
            broadcasting.
        parameters (None or dict): Parameters for the new
            #ak.contents.RecordArray node that is created by this operation.
        with_name (None or str): Assigns a `"__record__"` name to the new
            #ak.contents.RecordArray node that is created by this operation
            (overriding `parameters`, if necessary).
        right_broadcast (bool): If True, follow rules for implicit
            right-broadcasting, as described in #ak.broadcast_arrays.
        optiontype_outside_record (bool): If True, continue broadcasting past
            any option types before creating the new #ak.contents.RecordArray node.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Combines `arrays` into a single structure as the fields of a collection
    of records or the slots of a collection of tuples. If the `arrays` have
    nested structure, they are broadcasted with one another to form the
    records or tuples as deeply as possible, though this can be limited by
    `depth_limit`.

    This operation may be thought of as the opposite of projection in
    #ak.Array.__getitem__, which extracts fields one at a time, or
    #ak.unzip, which extracts them all in one call.

    Consider the following arrays, `one` and `two`.

        >>> one = ak.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6]])
        >>> two = ak.Array([["a", "b", "c"], [], ["d", "e"], ["f"]])

    Zipping them together using a dict creates a collection of records with
    the same nesting structure as `one` and `two`.

        >>> ak.zip({"x": one, "y": two}).show()
        [[{x: 1.1, y: 'a'}, {x: 2.2, y: 'b'}, {x: 3.3, y: 'c'}],
         [],
         [{x: 4.4, y: 'd'}, {x: 5.5, y: 'e'}],
         [{x: 6.6, y: 'f'}]]

    Doing so with a list creates tuples, whose fields are not named.

        >>> ak.zip([one, two]).show()
        [[(1.1, 'a'), (2.2, 'b'), (3.3, 'c')],
         [],
         [(4.4, 'd'), (5.5, 'e')],
         [(6.6, 'f')]]

    Adding a third array with the same length as `one` and `two` but less
    internal structure is okay: it gets broadcasted to match the others.
    (See #ak.broadcast_arrays for broadcasting rules.)

        >>> three = ak.Array([100, 200, 300, 400])
        >>> ak.zip([one, two, three]).show()
        [[(1.1, 'a', 100), (2.2, 'b', 100), (3.3, 'c', 100)],
         [],
         [(4.4, 'd', 300), (5.5, 'e', 300)],
         [(6.6, 'f', 400)]]

    However, if arrays have the same depth but different lengths of nested
    lists, attempting to zip them together is a broadcasting error.

        >>> one = ak.Array([[[1, 2, 3], [], [4, 5], [6]], [], [[7, 8]]])
        >>> two = ak.Array([[[1.1, 2.2], [3.3], [4.4], [5.5]], [], [[6.6]]])
        >>> ak.zip([one, two])
        ValueError: while calling
            ak.zip(
                arrays = [<Array [[[1, 2, 3], [], [4, ...], [6]], ...] type='3 * var ...
                depth_limit = None
                parameters = None
                with_name = None
                right_broadcast = False
                optiontype_outside_record = False
                highlevel = True
                behavior = None
            )
        Error details: cannot broadcast nested list

    For this, one can set the `depth_limit` to prevent the operation from
    attempting to broadcast what can't be broadcasted.

        >>> ak.zip([one, two], depth_limit=1).show()
        [([[1, 2, 3], [], [4, ...], [6]], [[1.1, ...], ...]),
         ([], []),
         ([[7, 8]], [[6.6]])]

    As an extreme, `depth_limit=1` is a handy way to make a record structure
    at the outermost level, regardless of whether the fields have matching
    structure or not.

    When zipping together arrays with optional values, it can be useful to create
    the #ak.contents.RecordArray node after the option types. By default, #ak.zip
    does not do this:

        >>> one = ak.Array([1, 2, None])
        >>> two = ak.Array([None, 5, 6])
        >>> ak.zip([one, two])
        <Array [(1, None), (2, 5), (None, 6)] type='3 * (?int64, ?int64)'>

    If the `optiontype_outside_record` option is set to `True`, Awkward will continue to
    broadcast the arrays together at the depth_limit until it reaches non-option
    types. This effectively takes the union of the option mask:

        >>> ak.zip([one, two], optiontype_outside_record=True)
        <Array [None, (2, 5), None] type='3 * ?(int64, int64)'>
    """
    # Dispatch
    if isinstance(arrays, Mapping):
        yield arrays.values()
    else:
        yield arrays

    # Implementation
    return _impl(
        arrays,
        depth_limit,
        parameters,
        with_name,
        right_broadcast,
        optiontype_outside_record,
        highlevel,
        behavior,
    )


def _impl(
    arrays,
    depth_limit,
    parameters,
    with_name,
    right_broadcast,
    optiontype_outside_record,
    highlevel,
    behavior,
):
    if depth_limit is not None and depth_limit <= 0:
        raise ValueError("depth_limit must be None or at least 1")

    if isinstance(arrays, dict):
        behavior = behavior_of(*arrays.values(), behavior=behavior)
        recordlookup = []
        layouts = []
        num_scalars = 0
        for n, x in arrays.items():
            recordlookup.append(n)
            try:
                layout = ak.operations.to_layout(
                    x, allow_record=False, allow_other=False
                )
            except TypeError:
                num_scalars += 1
                layout = ak.operations.to_layout(
                    [x], allow_record=False, allow_other=False
                )
            layouts.append(layout)

    else:
        arrays = list(arrays)
        behavior = behavior_of(*arrays, behavior=behavior)
        recordlookup = None
        layouts = []
        num_scalars = 0
        for x in arrays:
            try:
                layout = ak.operations.to_layout(
                    x, allow_record=False, allow_other=False
                )
            except TypeError:
                num_scalars += 1
                layout = ak.operations.to_layout(
                    [x], allow_record=False, allow_other=False
                )
            layouts.append(layout)

    to_record = num_scalars == len(arrays)

    if with_name is not None:
        if parameters is None:
            parameters = {}
        else:
            parameters = dict(parameters)
        parameters["__record__"] = with_name

    def action(inputs, depth, **ignore):
        if depth_limit == depth or all(
            x.purelist_depth == 1
            or (
                x.purelist_depth == 2
                and x.purelist_parameter("__array__") in ("string", "bytestring")
            )
            for x in inputs
        ):
            # If we want to zip after option types at this depth
            if optiontype_outside_record and any(x.is_option for x in inputs):
                return None

            return (
                ak.contents.RecordArray(inputs, recordlookup, parameters=parameters),
            )
        else:
            return None

    out = ak._broadcasting.broadcast_and_apply(
        layouts, action, behavior, right_broadcast=right_broadcast
    )
    assert isinstance(out, tuple) and len(out) == 1
    out = out[0]

    if to_record:
        out = out[0]
        assert isinstance(out, ak.record.Record)

    return wrap_layout(out, behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak

to_list = ak.operations.to_list


def test_1d():
    array = ak.operations.from_iter(
        [3.3, 2.2, 5.5, 1.1, 4.4],
        highlevel=False,
    )
    assert to_list(ak.argmin(array, axis=0, highlevel=False)) == 3
    assert to_list(ak.argmax(array, axis=0, highlevel=False)) == 2
    assert to_list(ak.count(array, axis=0, highlevel=False)) == 5
    assert to_list(ak.count_nonzero(array, axis=0, highlevel=False)) == 5
    assert to_list(ak.sum(array, axis=0, highlevel=False)) == 16.5
    assert to_list(ak.prod(array, axis=0, highlevel=False)) == 193.26120000000003
    assert to_list(ak.any(array, axis=0, highlevel=False)) is True
    assert to_list(ak.all(array, axis=0, highlevel=False)) is True
    assert to_list(ak.min(array, axis=0, highlevel=False)) == 1.1
    assert to_list(ak.max(array, axis=0, highlevel=False)) == 5.5


def test_2d():
    array = ak.operations.from_iter(
        [
            [3.3, 2.2, 5.5, 1.1, 4.4],
            [4.4, 2.2, 1.1, 3.3, 5.5],
            [2.2, 1.1, 4.4, 3.3, 5.5],
        ],
        highlevel=False,
    )
    assert to_list(ak.argmin(array, axis=0, highlevel=False)) == [2, 2, 1, 0, 0]
    assert (
        ak.argmin(array.to_typetracer(), axis=0, highlevel=False).form
        == ak.argmin(array, axis=0, highlevel=False).form
    )
    assert to_list(ak.argmin(array, axis=1, highlevel=False)) == [3, 2, 1]
    assert (
        ak.argmin(array.to_typetracer(), axis=1, highlevel=False).form
        == ak.argmin(array, axis=1, highlevel=False).form
    )


def test_3d():
    array = ak.operations.from_iter(
        [
            [
                [3.3, 2.2, 5.5, 1.1, 4.4],
                [4.4, 2.2, 1.1, 3.3, 5.5],
                [2.2, 1.1, 4.4, 3.3, 5.5],
            ],
            [
                [-3.3, 2.2, -5.5, 1.1, 4.4],
                [4.4, -2.2, 1.1, -3.3, 5.5],
                [2.2, 1.1, 4.4, 3.3, -5.5],
            ],
        ],
        highlevel=False,
    )

    assert to_list(ak.argmin(array, axis=0, highlevel=False)) == [
        [1, 0, 1, 0, 0],
        [0, 1, 0, 1, 0],
        [0, 0, 0, 0, 1],
    ]
    assert to_list(ak.argmin(array, axis=1, highlevel=False)) == [
        [2, 2, 1, 0, 0],
        [0, 1, 0, 1, 2],
    ]
    assert (
        ak.argmin(array.to_typetracer(), axis=1, highlevel=False).form
        == ak.argmin(array, axis=1, highlevel=False).form
    )
    assert to_list(ak.argmin(array, axis=2, highlevel=False)) == [[3, 2, 1], [2, 3, 4]]
    assert (
        ak.argmin(array.to_typetracer(), axis=2, highlevel=False).form
        == ak.argmin(array, axis=2, highlevel=False).form
    )
    assert to_list(ak.argmin(array, axis=-1, highlevel=False)) == [[3, 2, 1], [2, 3, 4]]
    assert (
        ak.argmin(array.to_typetracer(), axis=-1, highlevel=False).form
        == ak.argmin(array, axis=-1, highlevel=False).form
    )
    assert to_list(ak.argmin(array, axis=-2, highlevel=False)) == [
        [2, 2, 1, 0, 0],
        [0, 1, 0, 1, 2],
    ]
    assert (
        ak.argmin(array.to_typetracer(), axis=-2, highlevel=False).form
        == ak.argmin(array, axis=-2, highlevel=False).form
    )
    assert to_list(ak.argmin(array, axis=-3, highlevel=False)) == [
        [1, 0, 1, 0, 0],
        [0, 1, 0, 1, 0],
        [0, 0, 0, 0, 1],
    ]
    assert (
        ak.argmin(array.to_typetracer(), axis=-3, highlevel=False).form
        == ak.argmin(array, axis=-3, highlevel=False).form
    )


def test_jagged():
    v2_array = ak.operations.from_iter(
        [[2.2, 1.1, 3.3], [], [4.4, 5.5], [5.5], [-4.4, -5.5, -6.6]], highlevel=False
    )
    assert to_list(ak.argmin(v2_array, axis=1, highlevel=False)) == [1, None, 0, 0, 2]
    assert (
        ak.argmin(v2_array.to_typetracer(), axis=1, highlevel=False).form
        == ak.argmin(v2_array, axis=1, highlevel=False).form
    )

    index2 = ak.index.Index64(np.array([4, 3, 2, 1, 0], dtype=np.int64))
    v2_array2 = ak.contents.IndexedArray(index2, v2_array)
    assert to_list(ak.argmin(v2_array2, axis=1, highlevel=False)) == [2, 0, 0, None, 1]
    assert (
        ak.argmin(v2_array2.to_typetracer(), axis=1, highlevel=False).form
        == ak.argmin(v2_array2, axis=1, highlevel=False).form
    )

    index3 = ak.index.Index64(np.array([4, 3, -1, 4, 0], dtype=np.int64))
    v2_array2 = ak.contents.IndexedOptionArray(index3, v2_array)
    assert to_list(ak.argmin(v2_array2, axis=1, highlevel=False)) == [2, 0, None, 2, 1]
    assert (
        ak.argmin(v2_array2.to_typetracer(), axis=1, highlevel=False).form
        == ak.argmin(v2_array2, axis=1, highlevel=False).form
    )
    assert to_list(ak.argmin(v2_array2, axis=-1, highlevel=False)) == [2, 0, None, 2, 1]
    assert (
        ak.argmin(v2_array2.to_typetracer(), axis=-1, highlevel=False).form
        == ak.argmin(v2_array2, axis=-1, highlevel=False).form
    )


def test_missing():
    array = ak.operations.from_iter(
        [[[2.2, 1.1, 3.3]], [[]], [None, None, None], [[-4.4, -5.5, -6.6]]],
        highlevel=False,
    )
    assert to_list(ak.argmin(array, axis=2, highlevel=False)) == [
        [1],
        [None],
        [None, None, None],
        [2],
    ]
    assert (
        ak.argmin(array.to_typetracer(), axis=2, highlevel=False).form
        == ak.argmin(array, axis=2, highlevel=False).form
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
__all__ = ("concatenate",)
import awkward as ak
from awkward._backends.dispatch import backend_of
from awkward._backends.numpy import NumpyBackend
from awkward._behavior import behavior_of
from awkward._dispatch import high_level_function
from awkward._layout import maybe_posaxis, wrap_layout
from awkward._nplikes.numpylike import NumpyMetadata
from awkward._nplikes.shape import unknown_length
from awkward._regularize import regularize_axis
from awkward._typing import Sequence
from awkward.contents import Content
from awkward.operations.ak_fill_none import fill_none

np = NumpyMetadata.instance()
cpu = NumpyBackend.instance()


@ak._connect.numpy.implements("concatenate")
@high_level_function
def concatenate(arrays, axis=0, *, mergebool=True, highlevel=True, behavior=None):
    """
    Args:
        arrays: Array-like data (anything #ak.to_layout recognizes).
        axis (int): The dimension at which this operation is applied. The
            outermost dimension is `0`, followed by `1`, etc., and negative
            values count backward from the innermost: `-1` is the innermost
            dimension, `-2` is the next level up, etc.
        mergebool (bool): If True, boolean and numeric data can be combined
            into the same buffer, losing information about False vs `0` and
            True vs `1`; otherwise, they are kept in separate buffers with
            distinct types (using an #ak.contents.UnionArray).
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Returns an array with `arrays` concatenated. For `axis=0`, this means that
    one whole array follows another. For `axis=1`, it means that the `arrays`
    must have the same lengths and nested lists are each concatenated,
    element for element, and similarly for deeper levels.
    """
    # Dispatch
    if (
        # Is an array with a known backend
        backend_of(arrays, default=None)
        is not None
    ):
        yield (arrays,)
    else:
        yield arrays

    # Implementation
    return _impl(arrays, axis, mergebool, highlevel, behavior)


def _impl(arrays, axis, mergebool, highlevel, behavior):
    axis = regularize_axis(axis)
    # Simple single-array, axis=0 fast-path
    backend = backend_of(*arrays, default=cpu)
    behavior = behavior_of(*arrays, behavior=behavior)
    if (
        # Is an array with a known backend
        backend_of(arrays, default=None)
        is not None
    ):
        # Convert the array to a layout object
        content = ak.operations.to_layout(arrays, allow_record=False, allow_other=False)
        # Only handle concatenation along `axis=0`
        # Let ambiguous depth arrays fall through
        if maybe_posaxis(content, axis, 1) == 0:
            return ak.operations.ak_flatten._impl(content, 1, highlevel, behavior)

    content_or_others = [
        x.to_backend(backend) if isinstance(x, ak.contents.Content) else x
        for x in (
            ak.operations.to_layout(
                x, allow_record=False if axis == 0 else True, allow_other=True
            )
            for x in arrays
        )
    ]

    contents = [x for x in content_or_others if isinstance(x, ak.contents.Content)]
    if len(contents) == 0:
        raise ValueError("need at least one array to concatenate")

    posaxis = maybe_posaxis(contents[0], axis, 1)
    maxdepth = max(
        x.minmax_depth[1]
        for x in content_or_others
        if isinstance(x, ak.contents.Content)
    )
    if posaxis is None or not 0 <= posaxis < maxdepth:
        raise ValueError(
            "axis={} is beyond the depth of this array or the depth of this array "
            "is ambiguous".format(axis)
        )
    for x in content_or_others:
        if isinstance(x, ak.contents.Content):
            if maybe_posaxis(x, axis, 1) != posaxis:
                raise ValueError(
                    "arrays to concatenate do not have the same depth for negative "
                    "axis={}".format(axis)
                )

    if posaxis == 0:
        content_or_others = [
            x if isinstance(x, ak.contents.Content) else ak.operations.to_layout([x])
            for x in content_or_others
        ]
        batches = [[content_or_others[0]]]
        for x in content_or_others[1:]:
            batch = batches[-1]
            if ak._do.mergeable(batch[-1], x, mergebool=mergebool):
                batch.append(x)
            else:
                batches.append([x])

        contents = [ak._do.mergemany(b) for b in batches]
        if len(contents) > 1:
            out = _merge_as_union(contents)
        else:
            out = contents[0]

        if isinstance(out, ak.contents.UnionArray):
            out = type(out).simplified(
                out._tags,
                out._index,
                out._contents,
                parameters=out._parameters,
                mergebool=mergebool,
            )

    else:

        def action(inputs, depth, **kwargs):
            if any(
                x.minmax_depth == (1, 1)
                for x in inputs
                if isinstance(x, ak.contents.Content)
            ):
                raise ValueError(
                    "at least one array is not deep enough to concatenate at "
                    "axis={}".format(axis)
                )

            if depth != posaxis:
                return

            if any(isinstance(x, ak.contents.Content) and x.is_option for x in inputs):
                nextinputs = []
                for x in inputs:
                    if x.is_option and x.content.is_list:
                        nextinputs.append(fill_none(x, [], axis=0, highlevel=False))
                    else:
                        nextinputs.append(x)
                inputs = nextinputs

            # Ensure the lengths agree, taking known lengths over unknown lengths
            length = None
            for x in inputs:
                if isinstance(x, ak.contents.Content):
                    if length is None:
                        length = x.length
                    elif x.length is unknown_length:
                        continue
                    elif length is unknown_length:
                        length = x.length
                    elif length != x.length:
                        raise ValueError(
                            "all arrays must have the same length for "
                            "axis={}".format(axis)
                        )
            assert length is not None

            if all(
                (isinstance(x, ak.contents.Content) and x.is_regular)
                or (isinstance(x, ak.contents.NumpyArray) and x.data.ndim > 1)
                or not isinstance(x, ak.contents.Content)
                for x in inputs
            ):
                regulararrays = []
                sizes = []
                for x in inputs:
                    if isinstance(x, ak.contents.RegularArray):
                        regulararrays.append(x)
                    elif isinstance(x, ak.contents.NumpyArray):
                        regulararrays.append(x.to_RegularArray())
                    else:
                        regulararrays.append(
                            ak.contents.RegularArray(
                                ak.contents.NumpyArray(
                                    backend.nplike.broadcast_to(
                                        backend.nplike.asarray([x]), (length,)
                                    )
                                ),
                                1,
                            )
                        )
                    sizes.append(regulararrays[-1].size)

                prototype = backend.index_nplike.empty(sum(sizes), dtype=np.int8)
                start = 0
                for tag, size in enumerate(sizes):
                    prototype[start : start + size] = tag
                    start += size

                tags = ak.index.Index8(
                    backend.index_nplike.reshape(
                        backend.index_nplike.broadcast_to(
                            prototype, (length, prototype.size)
                        ),
                        (-1,),
                    )
                )
                index = ak.contents.UnionArray.regular_index(tags, backend=backend)
                inner = ak.contents.UnionArray.simplified(
                    tags,
                    index,
                    [x._content for x in regulararrays],
                    mergebool=mergebool,
                )

                return (ak.contents.RegularArray(inner, prototype.size),)

            elif all(
                isinstance(x, ak.contents.Content)
                and x.is_list
                or (isinstance(x, ak.contents.NumpyArray) and x.data.ndim > 1)
                or not isinstance(x, ak.contents.Content)
                for x in inputs
            ):
                nextinputs = []
                for x in inputs:
                    if isinstance(x, ak.contents.Content):
                        nextinputs.append(x)
                    else:
                        nextinputs.append(
                            ak.contents.ListOffsetArray(
                                ak.index.Index64(
                                    backend.index_nplike.arange(
                                        backend.index_nplike.shape_item_as_index(
                                            length + 1
                                        ),
                                        dtype=np.int64,
                                    ),
                                    nplike=backend.index_nplike,
                                ),
                                ak.contents.NumpyArray(
                                    backend.nplike.broadcast_to(
                                        backend.nplike.asarray([x]), (length,)
                                    )
                                ),
                            )
                        )

                counts = backend.index_nplike.zeros(
                    nextinputs[0].length, dtype=np.int64
                )
                all_counts = []
                all_flatten = []

                for x in nextinputs:
                    o, f = x._offsets_and_flattened(1, 1)
                    o = backend.index_nplike.asarray(o)
                    c = o[1:] - o[:-1]
                    backend.index_nplike.add(counts, c, maybe_out=counts)
                    all_counts.append(c)
                    all_flatten.append(f)

                offsets = backend.index_nplike.empty(
                    nextinputs[0].length + 1, dtype=np.int64
                )
                offsets[0] = 0
                backend.index_nplike.cumsum(counts, maybe_out=offsets[1:])

                offsets = ak.index.Index64(offsets, nplike=backend.index_nplike)

                tags, index = ak.contents.UnionArray.nested_tags_index(
                    offsets,
                    [ak.index.Index64(x) for x in all_counts],
                    backend=backend,
                )

                inner = ak.contents.UnionArray.simplified(
                    tags, index, all_flatten, mergebool=mergebool
                )

                return (ak.contents.ListOffsetArray(offsets, inner),)

            else:
                return None

        out = ak._broadcasting.broadcast_and_apply(
            content_or_others,
            action,
            behavior=behavior,
            allow_records=True,
            right_broadcast=False,
        )[0]

    return wrap_layout(out, behavior, highlevel)


def _merge_as_union(
    contents: Sequence[Content], parameters=None
) -> ak.contents.UnionArray:
    length = sum([c.length for c in contents])
    first = contents[0]
    tags = ak.index.Index8.empty(length, first.backend.index_nplike)
    index = ak.index.Index64.empty(length, first.backend.index_nplike)

    offset = 0
    for i, content in enumerate(contents):
        content.backend.maybe_kernel_error(
            content.backend["awkward_UnionArray_filltags_const", tags.dtype.type](
                tags.data, offset, content.length, i
            )
        )
        content.backend.maybe_kernel_error(
            content.backend["awkward_UnionArray_fillindex_count", index.dtype.type](
                index.data, offset, content.length
            )
        )
        offset += content.length

    return ak.contents.UnionArray.simplified(
        tags, index, contents, parameters=parameters
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
from __future__ import annotations

import numbers
import os
from collections.abc import Iterable, Sequence, Sized

from awkward._nplikes.numpylike import NumpyMetadata
from awkward._typing import AxisMaybeNone, SupportsInt

np = NumpyMetadata.instance()


def is_file_path(x) -> bool:
    try:
        return os.path.isfile(x)
    except ValueError:
        return False


def is_sized_iterable(obj) -> bool:
    return isinstance(obj, Iterable) and isinstance(obj, Sized)


def is_integer(x) -> bool:
    return isinstance(x, numbers.Integral) and not isinstance(x, bool)


def is_array_like(x) -> bool:
    return hasattr(x, "shape") and hasattr(x, "dtype") and hasattr(x, "T")


def is_integer_like(x) -> bool:
    # Integral types
    if isinstance(x, numbers.Integral):
        return not isinstance(x, bool)
    # Scalar arrays
    elif is_array_like(x):
        return np.issubdtype(x.dtype, np.integer) and x.ndim == 0
    # Other things that support integers
    else:
        return hasattr(x, "__int__")


def is_non_string_like_iterable(obj) -> bool:
    return not isinstance(obj, (str, bytes)) and isinstance(obj, Iterable)


def is_non_string_like_sequence(obj) -> bool:
    return not isinstance(obj, (str, bytes)) and isinstance(obj, Sequence)


def regularize_axis(axis: SupportsInt | None) -> AxisMaybeNone:
    if axis is None:
        return None
    else:
        return int(axis)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np  # noqa: F401
import pytest

import awkward as ak


def test():
    array = ak.Array([[0, 1, 2, 3], [8, 9, 10, 11]], backend="typetracer")
    other = ak.Array([1, 2], backend="cpu")
    result = array + other
    assert ak.backend(result) == "typetracer"


def test_mixed():
    pytest.importorskip("jax")
    ak.jax.register_and_check()

    array = ak.Array([[0, 1, 2, 3], [8, 9, 10, 11]], backend="cpu")
    other = ak.Array([1, 2], backend="jax")
    with pytest.raises(ValueError):
        array + other



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test_array():
    array = np.random.random(size=512).astype(dtype=np.float64)
    assert ak.type(array) == ak.types.ArrayType(ak.types.NumpyType("float64"), 512)


def test_dtype():
    assert ak.type(np.dtype(np.float64)) == ak.types.ScalarType(
        ak.types.NumpyType("float64")
    )


def test_type_instance():
    assert ak.type(np.float64(10.0)) == ak.types.ScalarType(
        ak.types.NumpyType("float64")
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import pytest

import awkward as ak


def test():
    array = ak.Array(
        ["is", "this", "real", "life", "or", "is", "this", "just", "fantasy"]
    )
    with pytest.raises(ValueError, match=r"strings cannot be directly flattened"):
        ak.flatten(array, axis=1)

    array = ak.Array(
        ["is", "this", "real", "life", "or", "is", "this", "just", "fantasy"]
    )
    assert ak.to_layout(array) == ak.flatten(array, axis=0, highlevel=False)

    array_ragged = ak.Array(
        [["is", "this", "real", "life"], ["or", "is", "this", "just", "fantasy"]]
    )
    assert ak.almost_equal(array, ak.flatten(array_ragged, axis=1))

    with pytest.raises(ValueError, match=r"strings cannot be directly flattened"):
        ak.flatten(array_ragged, axis=2)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
__all__ = ("count_nonzero",)
import awkward as ak
from awkward._behavior import behavior_of
from awkward._dispatch import high_level_function
from awkward._layout import wrap_layout
from awkward._nplikes.numpylike import NumpyMetadata
from awkward._regularize import regularize_axis

np = NumpyMetadata.instance()


@high_level_function
def count_nonzero(
    array,
    axis=None,
    *,
    keepdims=False,
    mask_identity=False,
    highlevel=True,
    behavior=None,
):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        axis (None or int): If None, combine all values from the array into
            a single scalar result; if an int, group by that axis: `0` is the
            outermost, `1` is the first level of nested lists, etc., and
            negative `axis` counts from the innermost: `-1` is the innermost,
            `-2` is the next level up, etc.
        keepdims (bool): If False, this reducer decreases the number of
            dimensions by 1; if True, the reduced values are wrapped in a new
            length-1 dimension so that the result of this operation may be
            broadcasted with the original array.
        mask_identity (bool): If True, reducing over empty lists results in
            None (an option type); otherwise, reducing over empty lists
            results in the operation's identity.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Counts nonzero elements of `array` (many types supported, including all
    Awkward Arrays and Records). The identity of counting is `0` and it is
    usually not masked. This operation is the same as NumPy's
    [count_nonzero](https://docs.scipy.org/doc/numpy/reference/generated/numpy.count_nonzero.html)
    if all lists at a given dimension have the same length and no None values,
    but it generalizes to cases where they do not.

    See #ak.sum for a more complete description of nested list and missing
    value (None) handling in reducers.

    Following the same rules as other reducers, #ak.count_nonzero does not
    count None values. If it is desirable to count them, use #ak.fill_none
    to turn them into something that would be counted.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, axis, keepdims, mask_identity, highlevel, behavior)


def _impl(array, axis, keepdims, mask_identity, highlevel, behavior):
    axis = regularize_axis(axis)
    layout = ak.operations.to_layout(array, allow_record=False, allow_other=False)
    behavior = behavior_of(array, behavior=behavior)
    reducer = ak._reducers.CountNonzero()

    out = ak._do.reduce(
        layout,
        reducer,
        axis=axis,
        mask=mask_identity,
        keepdims=keepdims,
        behavior=behavior,
    )
    if isinstance(out, (ak.contents.Content, ak.record.Record)):
        return wrap_layout(out, behavior, highlevel)
    else:
        return out


@ak._connect.numpy.implements("count_nonzero")
def _nep_18_impl(a, axis=None, *, keepdims=False):
    return count_nonzero(a, axis=axis, keepdims=keepdims)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import cupy as cp
import numpy as np
import pytest  # noqa: F401

import awkward as ak
from awkward._nplikes.cupy import Cupy
from awkward._nplikes.numpy import Numpy


def test_cupy_interop():
    cupy = Cupy.instance()
    numpy = Numpy.instance()

    cupy_index_arr = ak.index.Index64(cp.arange(10))
    assert cupy_index_arr.nplike is cupy

    np_index_arr = ak.index.Index64(np.arange(10))
    assert np_index_arr.nplike is numpy

    cupy_as_numpy = cupy_index_arr.to_nplike(numpy)
    assert isinstance(cupy_as_numpy.data, np.ndarray) and not isinstance(
        cupy_as_numpy.data, cp.ndarray
    )

    numpy_as_cupy = np_index_arr.to_nplike(cupy)
    assert isinstance(numpy_as_cupy.data, cp.ndarray)

    # GPU->CPU
    assert cupy_as_numpy.data.tolist() == np.asarray(np_index_arr).tolist()
    # CPU->CPU
    assert numpy_as_cupy.data.tolist() == np.asarray(np_index_arr).tolist()

    numpy_round_trip = numpy_as_cupy.to_nplike(numpy)
    assert isinstance(numpy_round_trip.data, np.ndarray)

    # CPU->GPU->CPU
    assert numpy_round_trip.data.tolist() == np.asarray(np_index_arr).tolist()



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
__all__ = ("to_rdataframe",)
from collections.abc import Mapping

import awkward as ak
from awkward._backends.numpy import NumpyBackend
from awkward._dispatch import high_level_function

cpu = NumpyBackend.instance()


@high_level_function
def to_rdataframe(arrays, *, flatlist_as_rvec=True):
    """
    Args:
        arrays (dict of arrays): Each value in this dict can be any array-like data
            that #ak.to_layout recognizes, but they must all have the same length.
        flatlist_as_rvec (bool): If True, lists of primitive types (numbers, booleans, etc.)
            are presented to C++ as `ROOT::RVec<primitive>`, but all other types use
            Awkward Array's custom C++ classes. If False, even these "flat" lists use
            Awkward Array's custom C++ classes.

    Converts an Awkward Array into ROOT Data Frame columns:

        >>> x = ak.Array([
        ...     [1.1, 2.2, 3.3],
        ...     [],
        ...     [4.4, 5.5],
        ... ])
        >>> y = ak.Array([
        ...     {"a": 1.1, "b": [1]},
        ...     {"a": 2.2, "b": [2, 1]},
        ...     {"a": 3.3, "b": [3, 2, 1]},
        ... ])

        >>> rdf = ak.to_rdataframe({"x": x, "y": y})
        >>> rdf.Define("z", "ROOT::VecOps::Sum(x) + y.a() + y.b()[0]").AsNumpy(["z"])
        {'z': ndarray([ 8.7,  4.2, 16.2])}

        >>> ak.sum(x, axis=-1) + y.a + y.b[:, 0]
        <Array [8.7, 4.2, 16.2] type='3 * float64'>

    See also #ak.from_rdataframe.
    """
    # Dispatch
    yield arrays.values()

    # Implementation
    return _impl(
        arrays,
        flatlist_as_rvec=flatlist_as_rvec,
    )


def _impl(
    arrays,
    flatlist_as_rvec,
):
    import awkward._connect.rdataframe.to_rdataframe  # noqa: F401

    if not isinstance(arrays, Mapping):
        raise TypeError("'arrays' must be a dict (to provide C++ names for the arrays)")
    elif not all(isinstance(name, str) for name in arrays):
        raise TypeError(
            "keys of 'arrays' dict must be strings (to provide C++ names for the arrays)"
        )
    elif len(arrays) == 0:
        raise TypeError("'arrays' must contain at least one array")

    layouts = {}
    length = None
    for name, array in arrays.items():
        layouts[name] = ak.operations.ak_to_layout._impl(
            array, allow_record=False, allow_other=False, regulararray=True
        ).to_backend(cpu)
        if length is None:
            length = layouts[name].length
        elif length != layouts[name].length:
            raise ValueError("lengths of 'arrays' must all be the same")

    return ak._connect.rdataframe.to_rdataframe.to_rdataframe(
        layouts,
        length,
        flatlist_as_rvec=flatlist_as_rvec,
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest

import awkward as ak


@pytest.mark.parametrize("string", ["hello world!", b"hello world!"])
def test(string):
    assert ak.almost_equal(
        ak.to_layout(string),
        ak.contents.NumpyArray(
            np.array(
                [104, 101, 108, 108, 111, 32, 119, 111, 114, 108, 100, 33],
                dtype=np.uint8,
            ),
            parameters={"__array__": "char" if isinstance(string, str) else "byte"},
        ),
    )


@pytest.mark.parametrize("string", ["hello world!", b"hello world!"])
def test_ufunc(string):
    array = ak.Array([string, string])
    assert ak.all(array == string)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak

tuple = ak.contents.RecordArray([ak.contents.NumpyArray(np.arange(10))], None)
tuple2 = ak.contents.RecordArray(
    [ak.from_iter([str(x) for x in range(10)], highlevel=False)], None
)
record = ak.contents.RecordArray([ak.contents.NumpyArray(np.arange(10))], ["x"])
record2 = ak.contents.RecordArray(
    [ak.from_iter([str(x) for x in range(10)], highlevel=False)], ["x"]
)


def test_record():
    array = ak.Array(record)

    assert not ak.is_tuple(array)


def test_tuple():
    array = ak.Array(tuple)

    assert ak.is_tuple(array)


def test_numpy():
    array = ak.Array(ak.contents.NumpyArray(np.arange(10)))

    assert not ak.is_tuple(array)


def test_list():
    array = ak.Array(
        ak.contents.ListArray(
            ak.index.Index64(np.array([0, 2], dtype=np.int64)),
            ak.index.Index64(np.array([2, 4], dtype=np.int64)),
            tuple,
        )
    )

    assert ak.is_tuple(array)

    array = ak.Array(
        ak.contents.ListArray(
            ak.index.Index64(np.array([0, 2], dtype=np.int64)),
            ak.index.Index64(np.array([2, 4], dtype=np.int64)),
            record,
        )
    )

    assert not ak.is_tuple(array)


def test_listoffset():
    array = ak.Array(
        ak.contents.ListOffsetArray(
            ak.index.Index64(np.array([0, 2, 4], dtype=np.int64)),
            tuple,
        )
    )

    assert ak.is_tuple(array)

    array = ak.Array(
        ak.contents.ListOffsetArray(
            ak.index.Index64(np.array([0, 2, 4], dtype=np.int64)), record
        )
    )

    assert not ak.is_tuple(array)


def test_indexed():
    array = ak.Array(
        ak.contents.IndexedArray(
            ak.index.Index64(np.array([0, 1, 3], dtype=np.int64)), tuple
        )
    )

    assert ak.is_tuple(array)

    array = ak.Array(
        ak.contents.IndexedArray(
            ak.index.Index64(np.array([0, 1, 3], dtype=np.int64)), record
        )
    )

    assert not ak.is_tuple(array)


def test_regular():
    array = ak.Array(ak.contents.RegularArray(tuple, 5))

    assert ak.is_tuple(array)

    array = ak.Array(ak.contents.RegularArray(record, 5))

    assert not ak.is_tuple(array)


def test_bytemasked():
    array = ak.Array(
        ak.contents.ByteMaskedArray(
            ak.index.Index8(np.array([0, 1, 0, 1], dtype=np.int64)),
            tuple,
            valid_when=True,
        )
    )

    assert ak.is_tuple(array)

    array = ak.Array(
        ak.contents.ByteMaskedArray(
            ak.index.Index8(np.array([0, 1, 0, 1], dtype=np.int64)),
            record,
            valid_when=True,
        )
    )

    assert not ak.is_tuple(array)


def test_bitmasked():
    array = ak.Array(
        ak.contents.BitMaskedArray(
            ak.index.IndexU8(np.array([0, 1, 0, 1], dtype=np.int64)),
            tuple,
            valid_when=True,
            length=4,
            lsb_order=True,
        )
    )

    assert ak.is_tuple(array)

    array = ak.Array(
        ak.contents.BitMaskedArray(
            ak.index.IndexU8(np.array([0, 1, 0, 1], dtype=np.int64)),
            record,
            valid_when=True,
            length=4,
            lsb_order=True,
        )
    )

    assert not ak.is_tuple(array)


def test_union():
    array = ak.Array(
        ak.contents.UnionArray(
            ak.index.Index8([0, 0, 1, 1]),
            ak.index.Index64([0, 1, 0, 1]),
            [tuple, ak.contents.NumpyArray(np.arange(10))],
        )
    )

    assert not ak.is_tuple(array)

    array = ak.Array(
        ak.contents.UnionArray(
            ak.index.Index8([0, 0, 1, 1]),
            ak.index.Index64([0, 1, 0, 1]),
            [tuple, tuple2],
        )
    )

    assert ak.is_tuple(array)

    array = ak.Array(
        ak.contents.UnionArray(
            ak.index.Index8([0, 0, 1, 1]),
            ak.index.Index64([0, 1, 0, 1]),
            [record, ak.contents.NumpyArray(np.arange(10))],
        )
    )

    assert not ak.is_tuple(array)

    array = ak.Array(
        ak.contents.UnionArray(
            ak.index.Index8([0, 0, 1, 1]),
            ak.index.Index64([0, 1, 0, 1]),
            [record, tuple],
        )
    )

    assert not ak.is_tuple(array)

    array = ak.Array(
        ak.contents.UnionArray(
            ak.index.Index8([0, 0, 1, 1]),
            ak.index.Index64([0, 1, 0, 1]),
            [record, record2],
        )
    )

    assert not ak.is_tuple(array)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
__all__ = ("metadata_from_parquet",)
import collections

import awkward as ak
from awkward._dispatch import high_level_function
from awkward._nplikes.numpylike import NumpyMetadata

np = NumpyMetadata.instance()


ParquetMetadata = collections.namedtuple(
    "ParquetMetadata",
    ["form", "fs", "paths", "metadata"],
)


@high_level_function
def metadata_from_parquet(
    path,
    *,
    storage_options=None,
    row_groups=None,
    ignore_metadata=False,
    scan_files=True,
):
    """
    Args:
        path (str): Local filename or remote URL, passed to fsspec for resolution.
            May contain glob patterns. A list of paths is also allowed, but they
            must be data files, not directories.
        storage_options: Passed to `fsspec.parquet.open_parquet_file`.
        row_groups (None or set of int): Row groups to read; must be non-negative.
            Order is ignored: the output array is presented in the order specified
            by Parquet metadata. If None, all row groups/all rows are read.
        ignore_metadata (bool): ignore the dedicated _metadata file if found
            and instead derive metadata from the first data file.
        scan_files (bool): TODO

    This function differs from ak.from_parquet._metadata as follows:

    * this function will always use a _metadata file, if present
    * if there is no _metadata, the schema comes from _common_metadata or
      the first data file
    * the total number of rows is always known

    Returns dict containing

    * `form`: an Awkward Form representing the low-level type of the data
      (use `.type` to get a high-level type),
    * `fs`: the fsspec filesystem object,
    * `paths`: a list of matching path names,
    * `col_counts`: the number of rows in each row group,
    * `columns`: the columns defined by the schema,
    * `num_rows`: the length of the array that would be read by #ak.from_parquet,
    * `num_row_groups`: the units that can be filtered (for the #ak.from_parquet `row_groups`
      argument).

    See also #ak.from_parquet, #ak.to_parquet.
    """
    import awkward._connect.pyarrow  # noqa: F401

    return _impl(
        path,
        storage_options,
        row_groups=row_groups,
        ignore_metadata=ignore_metadata,
        scan_files=scan_files,
    )


def _impl(
    path, storage_options, row_groups=None, ignore_metadata=False, scan_files=True
):
    results = ak.operations.ak_from_parquet.metadata(
        path, storage_options, row_groups, None, ignore_metadata, scan_files
    )
    parquet_columns, subform, actual_paths, fs, subrg, col_counts, metadata = results

    out = {
        "form": subform,
        "fs": fs,
        "paths": actual_paths,
        "col_counts": col_counts,
        "columns": parquet_columns,
    }
    if col_counts:
        out["num_rows"] = sum(col_counts)
        out["num_row_groups"] = len(col_counts)
    else:
        out["num_rows"] = None
        out["num_row_groups"] = None
    return out



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
import numpy as np

import awkward as ak


def test_tuple():
    data = (np.array([0, 1], dtype=np.int64), np.array([1, 0], dtype=np.int64))
    result = np.ravel_multi_index(
        (ak.from_numpy(data[0]), ak.from_numpy(data[1])), (2, 2)
    )
    assert isinstance(result, ak.Array)
    assert ak.almost_equal(result, np.ravel_multi_index(data, (2, 2)))


def test_list():
    A = np.eye(2) * 2
    B = np.eye(3) * 3
    result = np.block(
        [
            [ak.from_numpy(A), ak.from_numpy(np.zeros((2, 3)))],
            [ak.from_numpy(np.ones((3, 2))), ak.from_numpy(B)],
        ]
    )
    assert isinstance(result, ak.Array)
    assert ak.almost_equal(
        result, np.block([[A, np.zeros((2, 3))], [np.ones((3, 2)), B]])
    )


def test_array():
    haystack = np.array([1, 2, 3, 4, 4, 5, 6, 7], dtype=np.int64)
    needle = np.array([5, 0, 2], dtype=np.int64)
    result = np.searchsorted(ak.from_numpy(haystack), ak.from_numpy(needle))
    assert isinstance(result, ak.Array)
    assert ak.almost_equal(result, np.searchsorted(haystack, needle))


def test_scalar():
    data = np.array([1, 2, 3, 4, 3, 2, 1, 2], dtype=np.int64)
    result = np.partition(ak.from_numpy(data), 4)
    assert isinstance(result, ak.Array)
    assert ak.almost_equal(result, np.partition(data, 4))


def test_tuple_of_array():
    data = (
        np.array([5, 4, 3, 2, 1], dtype=np.int64),
        np.array([1, 2, 3, 5, 0], dtype=np.int64),
    )
    result = np.lexsort((ak.from_numpy(data[0]), ak.from_numpy(data[1])))
    assert isinstance(result, ak.Array)
    assert ak.almost_equal(result, np.lexsort(data))



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest

import awkward as ak

to_list = ak.operations.to_list


def test_same_categories():
    categories = ak.Array(["one", "two", "three"])
    index1 = ak.index.Index64(np.array([0, 2, 2, 1, 2, 0, 1, 0], dtype=np.int64))
    index2 = ak.index.Index64(np.array([1, 1, 2, 1, 0, 0, 0, 1], dtype=np.int64))
    categorical1 = ak.contents.IndexedArray(
        index1, categories.layout, parameters={"__array__": "categorical"}
    )
    categorical2 = ak.contents.IndexedArray(
        index2, categories.layout, parameters={"__array__": "categorical"}
    )
    array1 = ak.Array(categorical1)
    array2 = ak.Array(categorical2)
    assert (array1 == array2).to_list() == [
        False,
        False,
        True,
        True,
        False,
        True,
        False,
        False,
    ]


def test_different_categories():
    categories1 = ak.Array(["one", "two", "three"])
    categories2 = ak.Array(["three", "two", "one"])
    index1 = ak.index.Index64(np.array([0, 2, 2, 1, 2, 0, 1, 0], dtype=np.int64))
    index2 = ak.index.Index64(np.array([1, 1, 2, 1, 0, 0, 0, 1], dtype=np.int64))
    categorical1 = ak.contents.IndexedArray(
        index1, categories1.layout, parameters={"__array__": "categorical"}
    )
    categorical2 = ak.contents.IndexedArray(
        index2, categories2.layout, parameters={"__array__": "categorical"}
    )
    array1 = ak.Array(categorical1)
    array2 = ak.Array(categorical2)
    assert (array1 == array2).to_list() == [
        False,
        False,
        False,
        True,
        True,
        False,
        False,
        False,
    ]


def test_one_extra():
    categories1 = ak.Array(["one", "two", "three", "four"])
    categories2 = ak.Array(["three", "two", "one"])
    index1 = ak.index.Index64(np.array([0, 3, 0, 1, 2, 3, 1, 0], dtype=np.int64))
    index2 = ak.index.Index64(np.array([1, 1, 2, 1, 0, 0, 0, 1], dtype=np.int64))
    categorical1 = ak.contents.IndexedArray(
        index1, categories1.layout, parameters={"__array__": "categorical"}
    )
    categorical2 = ak.contents.IndexedArray(
        index2, categories2.layout, parameters={"__array__": "categorical"}
    )
    array1 = ak.Array(categorical1)
    array2 = ak.Array(categorical2)
    assert (array1 == array2).to_list() == [
        False,
        False,
        True,
        True,
        True,
        False,
        False,
        False,
    ]


def test_two_extra():
    categories1 = ak.Array(["one", "two", "three"])
    categories2 = ak.Array(["four", "three", "two", "one"])
    index1 = ak.index.Index64(np.array([0, 2, 2, 1, 2, 0, 1, 0], dtype=np.int64))
    index2 = ak.index.Index64(np.array([1, 3, 1, 1, 3, 3, 2, 1], dtype=np.int64))
    categorical1 = ak.contents.IndexedArray(
        index1, categories1.layout, parameters={"__array__": "categorical"}
    )
    categorical2 = ak.contents.IndexedArray(
        index2, categories2.layout, parameters={"__array__": "categorical"}
    )
    array1 = ak.Array(categorical1)
    array2 = ak.Array(categorical2)
    assert (array1 == array2).to_list() == [
        False,
        False,
        True,
        False,
        False,
        True,
        True,
        False,
    ]


def test_option_same_categories():
    categories = ak.Array(["one", "two", "three"])
    index1 = ak.index.Index64(np.array([0, 2, 2, 1, -1, 0, 1, 0], dtype=np.int64))
    index2 = ak.index.Index64(np.array([1, 1, 2, 1, -1, -1, 0, 1], dtype=np.int64))
    categorical1 = ak.contents.IndexedOptionArray(
        index1, categories.layout, parameters={"__array__": "categorical"}
    )
    categorical2 = ak.contents.IndexedOptionArray(
        index2, categories.layout, parameters={"__array__": "categorical"}
    )
    array1 = ak.Array(categorical1)
    array2 = ak.Array(categorical2)
    assert (array1 == array2).to_list() == [
        False,
        False,
        True,
        True,
        True,
        False,
        False,
        False,
    ]


def test_option_different_categories():
    categories1 = ak.Array(["one", "two", "three"])
    categories2 = ak.Array(["three", "two", "one"])
    index1 = ak.index.Index64(np.array([0, 2, 2, 1, -1, -1, 1, 0], dtype=np.int64))
    index2 = ak.index.Index64(np.array([1, 1, 2, 1, 0, -1, 0, 1], dtype=np.int64))
    categorical1 = ak.contents.IndexedOptionArray(
        index1, categories1.layout, parameters={"__array__": "categorical"}
    )
    categorical2 = ak.contents.IndexedOptionArray(
        index2, categories2.layout, parameters={"__array__": "categorical"}
    )
    array1 = ak.Array(categorical1)
    array2 = ak.Array(categorical2)
    assert (array1 == array2).to_list() == [
        False,
        False,
        False,
        True,
        False,
        True,
        False,
        False,
    ]


def test_option_one_extra():
    categories1 = ak.Array(["one", "two", "three", "four"])
    categories2 = ak.Array(["three", "two", "one"])
    index1 = ak.index.Index64(np.array([0, -1, 0, 1, 2, 3, 1, 0], dtype=np.int64))
    index2 = ak.index.Index64(np.array([1, -1, -1, 1, 0, 0, 0, 1], dtype=np.int64))
    categorical1 = ak.contents.IndexedOptionArray(
        index1, categories1.layout, parameters={"__array__": "categorical"}
    )
    categorical2 = ak.contents.IndexedOptionArray(
        index2, categories2.layout, parameters={"__array__": "categorical"}
    )
    array1 = ak.Array(categorical1)
    array2 = ak.Array(categorical2)
    assert (array1 == array2).to_list() == [
        False,
        True,
        False,
        True,
        True,
        False,
        False,
        False,
    ]


def test_option_two_extra():
    categories1 = ak.Array(["one", "two", "three"])
    categories2 = ak.Array(["four", "three", "two", "one"])
    index1 = ak.index.Index64(np.array([0, -1, -1, 1, 2, 0, 1, 0], dtype=np.int64))
    index2 = ak.index.Index64(np.array([1, -1, 1, 1, 3, 3, 2, 1], dtype=np.int64))
    categorical1 = ak.contents.IndexedOptionArray(
        index1, categories1.layout, parameters={"__array__": "categorical"}
    )
    categorical2 = ak.contents.IndexedOptionArray(
        index2, categories2.layout, parameters={"__array__": "categorical"}
    )
    array1 = ak.Array(categorical1)
    array2 = ak.Array(categorical2)
    assert (array1 == array2).to_list() == [
        False,
        True,
        False,
        False,
        False,
        True,
        True,
        False,
    ]


def test_to_categorical_numbers():
    array = ak.Array([1.1, 2.2, 3.3, 1.1, 2.2, 3.3, 1.1, 2.2, 3.3])
    assert not ak.operations.ak_is_categorical.is_categorical(array)
    categorical = ak.operations.ak_to_categorical.to_categorical(array)
    assert ak.operations.ak_is_categorical.is_categorical(categorical)
    assert to_list(array) == categorical.to_list()
    assert to_list(categorical.layout.content) == [1.1, 2.2, 3.3]
    not_categorical = ak.operations.ak_from_categorical.from_categorical(categorical)
    assert not ak.operations.ak_is_categorical.is_categorical(not_categorical)
    assert ak.operations.ak_categories.categories(categorical).to_list() == [
        1.1,
        2.2,
        3.3,
    ]


def test_to_categorical_nested():
    array = ak.Array([["one", "two", "three"], [], ["one", "two"], ["three"]])
    assert not ak.operations.ak_is_categorical.is_categorical(array)
    categorical = ak.operations.ak_to_categorical.to_categorical(array)
    assert ak.operations.ak_is_categorical.is_categorical(categorical)
    assert to_list(array) == categorical.to_list()
    not_categorical = ak.operations.ak_from_categorical.from_categorical(categorical)
    assert not ak.operations.ak_is_categorical.is_categorical(not_categorical)
    assert ak.operations.ak_categories.categories(categorical).to_list() == [
        "one",
        "two",
        "three",
    ]


def test_to_categorical():
    array = ak.Array(
        ["one", "two", "three", "one", "two", "three", "one", "two", "three"]
    )
    assert not ak.operations.ak_is_categorical.is_categorical(array)
    categorical = ak.operations.ak_to_categorical.to_categorical(array)
    assert ak.operations.ak_is_categorical.is_categorical(categorical)
    assert to_list(array) == categorical.to_list()
    assert to_list(categorical.layout.content) == ["one", "two", "three"]
    not_categorical = ak.operations.ak_from_categorical.from_categorical(categorical)
    assert not ak.operations.ak_is_categorical.is_categorical(not_categorical)
    assert ak.operations.ak_categories.categories(categorical).to_list() == [
        "one",
        "two",
        "three",
    ]


def test_to_categorical_none():
    array = ak.Array(
        [
            "one",
            "two",
            "three",
            None,
            "one",
            "two",
            "three",
            None,
            "one",
            "two",
            "three",
            None,
        ]
    )
    assert not ak.operations.ak_is_categorical.is_categorical(array)
    categorical = ak.operations.ak_to_categorical.to_categorical(array)
    assert ak.operations.ak_is_categorical.is_categorical(categorical)
    assert to_list(array) == categorical.to_list()
    assert to_list(categorical.layout.content) == ["one", "two", "three"]
    not_categorical = ak.operations.ak_from_categorical.from_categorical(categorical)
    assert not ak.operations.ak_is_categorical.is_categorical(not_categorical)
    assert ak.operations.ak_categories.categories(categorical).to_list() == [
        "one",
        "two",
        "three",
    ]


def test_to_categorical_masked():
    content = ak.Array(
        [
            "one",
            "two",
            "three",
            "one",
            "one",
            "two",
            "three",
            "two",
            "one",
            "two",
            "three",
            "three",
        ]
    ).layout
    mask = ak.index.Index8(
        np.array(
            [
                False,
                False,
                False,
                True,
                False,
                False,
                False,
                True,
                False,
                False,
                False,
                True,
            ]
        )
    )
    array = ak.Array(ak.contents.ByteMaskedArray(mask, content, valid_when=False))
    assert not ak.operations.ak_is_categorical.is_categorical(array)
    categorical = ak.operations.ak_to_categorical.to_categorical(array)
    assert ak.operations.ak_is_categorical.is_categorical(categorical)
    assert to_list(array) == categorical.to_list()
    assert to_list(categorical.layout.content) == ["one", "two", "three"]
    not_categorical = ak.operations.ak_from_categorical.from_categorical(categorical)
    assert not ak.operations.ak_is_categorical.is_categorical(not_categorical)
    assert ak.operations.ak_categories.categories(categorical).to_list() == [
        "one",
        "two",
        "three",
    ]


def test_to_categorical_masked_again():
    content = ak.Array(
        ["one", "two", "three", "one", "one", "two", "three", "two"]
    ).layout
    index = ak.index.Index64(
        np.array([0, 1, 2, 3, 4, 5, 6, 7, 0, 1, 2, 3], dtype=np.int64)
    )
    indexedarray = ak.contents.IndexedArray(index, content)
    mask = ak.index.Index8(
        np.array(
            [
                False,
                False,
                False,
                True,
                False,
                False,
                False,
                True,
                False,
                False,
                False,
                True,
            ]
        )
    )
    array = ak.Array(
        ak.contents.ByteMaskedArray.simplified(mask, indexedarray, valid_when=False)
    )
    assert not ak.operations.ak_is_categorical.is_categorical(array)
    categorical = ak.operations.ak_to_categorical.to_categorical(array)
    assert ak.operations.ak_is_categorical.is_categorical(categorical)
    assert to_list(array) == categorical.to_list()
    assert to_list(categorical.layout.content) == ["one", "two", "three"]
    not_categorical = ak.operations.ak_from_categorical.from_categorical(categorical)
    assert not ak.operations.ak_is_categorical.is_categorical(not_categorical)
    assert ak.operations.ak_categories.categories(categorical).to_list() == [
        "one",
        "two",
        "three",
    ]


@pytest.mark.skip(reason="Fix issues for categorical type")
def test_typestr():
    assert (
        str(
            ak.operations.type(
                ak.operations.ak_to_categorical.to_categorical(
                    ak.Array([1.1, 2.2, 2.2, 3.3])
                )
            )
        )
        == "4 * categorical[type=float64]"
    )
    assert (
        str(
            ak.operations.type(
                ak.operations.ak_to_categorical.to_categorical(
                    ak.Array([1.1, 2.2, None, 2.2, 3.3])
                )
            )
        )
        == "5 * categorical[type=?float64]"
    )
    assert (
        str(
            ak.operations.type(
                ak.operations.ak_to_categorical.to_categorical(
                    ak.Array(["one", "two", "two", "three"])
                )
            )
        )
        == "4 * categorical[type=string]"
    )
    assert (
        str(
            ak.operations.type(
                ak.operations.ak_to_categorical.to_categorical(
                    ak.Array(["one", "two", None, "two", "three"])
                )
            )
        )
        == "5 * categorical[type=?string]"
    )


def test_zip():
    x = ak.Array([1.1, 2.2, 3.3])
    y = ak.Array(["one", "two", "three"])
    assert ak.zip({"x": x, "y": y}).to_list() == [
        {"x": 1.1, "y": "one"},
        {"x": 2.2, "y": "two"},
        {"x": 3.3, "y": "three"},
    ]
    y = ak.operations.ak_to_categorical.to_categorical(y)
    assert ak.zip({"x": x, "y": y}).to_list() == [
        {"x": 1.1, "y": "one"},
        {"x": 2.2, "y": "two"},
        {"x": 3.3, "y": "three"},
    ]


pyarrow = pytest.importorskip("pyarrow")


def test_arrow_nomask():
    array = ak.Array([1.1, 2.2, 3.3, 4.4, None])
    assert (
        str(ak.operations.type(ak.operations.from_arrow(ak.operations.to_arrow(array))))
        == "5 * ?float64"
    )
    assert (
        str(
            ak.operations.type(
                ak.operations.from_arrow(ak.operations.to_arrow(array[:-1]))
            )
        )
        == "4 * ?float64"
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak

to_list = ak.operations.to_list


def test_1406issue():
    array = ak.Array(
        ak.contents.ListOffsetArray(
            ak.index.Index64(np.array([1, 3], dtype=np.int64)),
            ak.contents.ListOffsetArray(
                ak.index.Index64(np.array([0, 2, 2, 3], dtype=np.int64)),
                ak.contents.NumpyArray(np.array([0, 1, 2], dtype=np.int64)),
            ),
        ),
        check_valid=True,
    )

    index = ak.Array(
        ak.contents.ListOffsetArray(
            ak.index.Index64(np.array([0, 2], dtype=np.int64)),
            ak.contents.IndexedOptionArray(
                ak.index.Index64(np.array([0, 1], dtype=np.int64)),
                ak.contents.ListOffsetArray(
                    ak.index.Index64(np.array([0, 0, 1], dtype=np.int64)),
                    ak.contents.NumpyArray(np.array([0], dtype=np.int64)),
                ),
            ),
        ),
        check_valid=True,
    )
    assert to_list(array[index]) == [[[], [2]]]


def test_success_remove_option_type():
    array = ak.Array(
        ak.contents.ListOffsetArray(
            ak.index.Index64(np.array([1, 3], dtype=np.int64)),
            ak.contents.ListOffsetArray(
                ak.index.Index64(np.array([0, 2, 2, 3], dtype=np.int64)),
                ak.contents.NumpyArray(np.array([0, 1, 2], dtype=np.int64)),
            ),
        ),
        check_valid=True,
    )

    index = ak.Array(
        ak.contents.ListOffsetArray(
            ak.index.Index64(np.array([0, 2], dtype=np.int64)),
            ak.contents.ListOffsetArray(
                ak.index.Index64(np.array([0, 0, 1], dtype=np.int64)),
                ak.contents.NumpyArray(np.array([0], dtype=np.int64)),
            ),
        ),
        check_valid=True,
    )

    assert to_list(array[index]) == [[[], [2]]]


def test_success_start_offset0():
    array = ak.Array(
        ak.contents.ListOffsetArray(
            ak.index.Index64(np.array([0, 2], dtype=np.int64)),
            ak.contents.ListOffsetArray(
                ak.index.Index64(np.array([2, 2, 3], dtype=np.int64)),
                ak.contents.NumpyArray(np.array([0, 1, 2], dtype=np.int64)),
            ),
        ),
        check_valid=True,
    )

    index = ak.Array(
        ak.contents.ListOffsetArray(
            ak.index.Index64(np.array([0, 2], dtype=np.int64)),
            ak.contents.IndexedOptionArray(
                ak.index.Index64(np.array([0, 1], dtype=np.int64)),
                ak.contents.ListOffsetArray(
                    ak.index.Index64(np.array([0, 0, 1], dtype=np.int64)),
                    ak.contents.NumpyArray(np.array([0], dtype=np.int64)),
                ),
            ),
        ),
        check_valid=True,
    )

    assert to_list(array[index]) == [[[], [2]]]


def test_success_nonempty_list():
    array = ak.Array(
        ak.contents.ListOffsetArray(
            ak.index.Index64(np.array([1, 3], dtype=np.int64)),
            ak.contents.ListOffsetArray(
                ak.index.Index64(np.array([0, 1, 2, 3], dtype=np.int64)),
                ak.contents.NumpyArray(np.array([0, 1, 2], dtype=np.int64)),
            ),
        ),
        check_valid=True,
    )

    index = ak.Array(
        ak.contents.ListOffsetArray(
            ak.index.Index64(np.array([0, 2], dtype=np.int64)),
            ak.contents.IndexedOptionArray(
                ak.index.Index64(np.array([0, 1], dtype=np.int64)),
                ak.contents.ListOffsetArray(
                    ak.index.Index64(np.array([0, 1, 2], dtype=np.int64)),
                    ak.contents.NumpyArray(np.array([0, 0], dtype=np.int64)),
                ),
            ),
        ),
        check_valid=True,
    )

    assert to_list(array[index]) == [[[1], [2]]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
__all__ = ("run_lengths",)
import awkward as ak
from awkward._backends.dispatch import backend_of
from awkward._backends.numpy import NumpyBackend
from awkward._behavior import behavior_of
from awkward._dispatch import high_level_function
from awkward._layout import wrap_layout
from awkward._nplikes.numpylike import NumpyMetadata
from awkward._nplikes.shape import unknown_length

np = NumpyMetadata.instance()
cpu = NumpyBackend.instance()


@high_level_function
def run_lengths(array, *, highlevel=True, behavior=None):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Computes the lengths of sequences of identical values at the deepest level
    of nesting, returning an array with the same structure but with `int64` type.

    For example,

        >>> array = ak.Array([1.1, 1.1, 1.1, 2.2, 3.3, 3.3, 4.4, 4.4, 5.5])
        >>> ak.run_lengths(array)
        <Array [3, 1, 2, 2, 1] type='5 * int64'>

    There are 3 instances of 1.1, followed by 1 instance of 2.2, 2 instances of 3.3,
    2 instances of 4.4, and 1 instance of 5.5.

    The order and uniqueness of the input data doesn't matter,

        >>> array = ak.Array([1.1, 1.1, 1.1, 5.5, 4.4, 4.4, 1.1, 1.1, 5.5])
        >>> ak.run_lengths(array)
        <Array [3, 1, 2, 2, 1] type='5 * int64'>

    just the difference between each value and its neighbors.

    The data can be nested, but runs don't cross list boundaries.

        >>> array = ak.Array([[1.1, 1.1, 1.1, 2.2, 3.3], [3.3, 4.4], [4.4, 5.5]])
        >>> ak.run_lengths(array)
        <Array [[3, 1, 1], [1, 1], [1, 1]] type='3 * var * int64'>

    This function recognizes strings as distinguishable values.

        >>> array = ak.Array([["one", "one"], ["one", "two", "two"], ["three", "two", "two"]])
        >>> ak.run_lengths(array)
        <Array [[2], [1, 2], [1, 2]] type='3 * var * int64'>

    Note that this can be combined with #ak.argsort and #ak.unflatten to compute
    a "group by" operation:

        >>> array = ak.Array([{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 1, "y": 1.1},
        ...                   {"x": 3, "y": 3.3}, {"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}])
        >>> sorted = array[ak.argsort(array.x)]
        >>> sorted.x
        <Array [1, 1, 1, 2, 2, 3] type='6 * int64'>
        >>> ak.run_lengths(sorted.x)
        <Array [3, 2, 1] type='3 * int64'>
        >>> ak.unflatten(sorted, ak.run_lengths(sorted.x)).show()
        [[{x: 1, y: 1.1}, {x: 1, y: 1.1}, {x: 1, y: 1.1}],
         [{x: 2, y: 2.2}, {x: 2, y: 2.2}],
         [{x: 3, y: 3.3}]]

    Unlike a database "group by," this operation can be applied in bulk to many sublists
    (though the run lengths need to be fully flattened to be used as `counts` for
    #ak.unflatten, and you need to specify `axis=-1` as the depth).

        >>> array = ak.Array([[{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 1, "y": 1.1}],
        ...                   [{"x": 3, "y": 3.3}, {"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}]])
        >>> sorted = array[ak.argsort(array.x)]
        >>> sorted.x
        <Array [[1, 1, 2], [1, 2, 3]] type='2 * var * int64'>
        >>> ak.run_lengths(sorted.x)
        <Array [[2, 1], [1, 1, 1]] type='2 * var * int64'>
        >>> counts = ak.flatten(ak.run_lengths(sorted.x), axis=None)
        >>> ak.unflatten(sorted, counts, axis=-1).show()
        [[[{x: 1, y: 1.1}, {x: 1, y: 1.1}], [{x: 2, y: 2.2}]],
         [[{x: 1, y: 1.1}], [{x: 2, y: 2.2}], [{x: 3, y: 3.3}]]]

    See also #ak.num, #ak.argsort, #ak.unflatten.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, highlevel, behavior)


def _impl(array, highlevel, behavior):
    backend = backend_of(array, default=cpu)

    def lengths_of(data, offsets):
        if backend.nplike.is_own_array(data):
            size = data.size
        else:
            layout = ak.to_layout(data)
            size = layout.length

        if size is not unknown_length and size == 0:
            return backend.index_nplike.empty(0, dtype=np.int64), offsets
        else:
            diffs = backend.nplike.asarray(data[1:] != data[:-1])
            # Do we have list boundaries to consider?
            if offsets is not None:
                # When checking to see whether one element equals its following neighbour
                # we also want to break runs at list boundaries. The comparison for offset `i`
                # occurs at `i-1` in `diffs`. Consider this example with an empty sublist:
                #                                data = [1 1 2 2 2 3 4 4 5  ]
                #                             offsets = [0           6   9 9]
                #                        (data) diffs = [  0 1 0 0 1 1 0 1  ]
                #                        diffs index  = [  0 1 2 3 4 5 6 7  ]
                #                                      boundary diff ^
                # To consider only the interior boundaries, we ignore the start and end
                # offset values. These can be repeated with empty sublists, so we mask them out.
                is_interior = backend.nplike.logical_and(
                    0 < offsets,
                    offsets < backend.index_nplike.shape_item_as_index(size),
                )
                interior_offsets = offsets[is_interior]
                diffs[interior_offsets - 1] = True
            positions = backend.index_nplike.nonzero(diffs)[0]
            full_positions = backend.index_nplike.empty(
                positions.size + 2, dtype=np.int64
            )
            full_positions[0] = 0
            full_positions[-1] = backend.index_nplike.shape_item_as_index(size)
            full_positions[1:-1] = positions + 1

            nextcontent = full_positions[1:] - full_positions[:-1]
            if offsets is None:
                nextoffsets = None
            else:
                nextoffsets = backend.index_nplike.searchsorted(
                    full_positions, offsets, side="left"
                )
            return nextcontent, nextoffsets

    def action(layout, **kwargs):
        if layout.branch_depth == (False, 1):
            if layout.is_indexed:
                layout = layout.project()

            if (
                layout.parameter("__array__") == "string"
                or layout.parameter("__array__") == "bytestring"
            ):
                nextcontent, _ = lengths_of(ak.highlevel.Array(layout), None)
                return ak.contents.NumpyArray(nextcontent)

            if not isinstance(layout, (ak.contents.NumpyArray, ak.contents.EmptyArray)):
                raise NotImplementedError("run_lengths on " + type(layout).__name__)

            nextcontent, _ = lengths_of(backend.nplike.asarray(layout), None)
            return ak.contents.NumpyArray(nextcontent)

        elif layout.branch_depth == (False, 2):
            if layout.is_indexed:
                layout = layout.project()

            if not layout.is_list:
                raise NotImplementedError("run_lengths on " + type(layout).__name__)

            if (
                layout.content.parameter("__array__") == "string"
                or layout.content.parameter("__array__") == "bytestring"
            ):
                # We also want to trim the _upper_ bound of content,
                # so we manually convert the list type to zero-based
                listoffsetarray = layout.to_ListOffsetArray64(False)
                offsets = backend.index_nplike.asarray(listoffsetarray.offsets)
                content = listoffsetarray.content[offsets[0] : offsets[-1]]

                if content.is_indexed:
                    content = content.project()

                nextcontent, nextoffsets = lengths_of(
                    ak.highlevel.Array(content), offsets - offsets[0]
                )
                return ak.contents.ListOffsetArray(
                    ak.index.Index64(nextoffsets),
                    ak.contents.NumpyArray(nextcontent),
                )

            listoffsetarray = layout.to_ListOffsetArray64(False)
            offsets = backend.index_nplike.asarray(listoffsetarray.offsets)
            content = listoffsetarray.content[offsets[0] : offsets[-1]]

            if content.is_indexed:
                content = content.project()

            if not isinstance(
                content, (ak.contents.NumpyArray, ak.contents.EmptyArray)
            ):
                raise NotImplementedError(
                    "run_lengths on "
                    + type(layout).__name__
                    + " with content "
                    + type(content).__name__
                )

            nextcontent, nextoffsets = lengths_of(
                backend.nplike.asarray(content), offsets - offsets[0]
            )
            return ak.contents.ListOffsetArray(
                ak.index.Index64(nextoffsets),
                ak.contents.NumpyArray(nextcontent),
            )
        else:
            return None

    layout = ak.operations.to_layout(array, allow_record=False, allow_other=False)
    behavior = behavior_of(array, behavior=behavior)

    out = ak._do.recursively_apply(layout, action, behavior)
    return wrap_layout(out, behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
__all__ = ("fill_none",)
import numbers

import awkward as ak
from awkward._backends.dispatch import backend_of
from awkward._backends.numpy import NumpyBackend
from awkward._behavior import behavior_of
from awkward._dispatch import high_level_function
from awkward._errors import AxisError
from awkward._layout import maybe_posaxis, wrap_layout
from awkward._nplikes.numpylike import NumpyMetadata
from awkward._regularize import is_sized_iterable, regularize_axis

np = NumpyMetadata.instance()
cpu = NumpyBackend.instance()


@high_level_function
def fill_none(array, value, axis=-1, *, highlevel=True, behavior=None):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        value: Data with which to replace None.
        axis (None or int): If None, replace all None values in the array
            with the given value; if an int, The dimension at which this
            operation is applied. The outermost dimension is `0`, followed
            by `1`, etc., and negative values count backward from the
            innermost: `-1` is the innermost  dimension, `-2` is the next
            level up, etc.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Replaces missing values (None) with a given `value`.

    For example, in the following

        >>> array = ak.Array([[1.1, None, 2.2], [], [None, 3.3, 4.4]])

    The None values could be replaced with `0` by

        >>> ak.fill_none(array, 0)
        <Array [[1.1, 0, 2.2], [], [0, 3.3, 4.4]] type='3 * var * float64'>

    The replacement value doesn't strictly need the same type as the
    surrounding data. For example, the None values could also be replaced
    by a string.

        >>> ak.fill_none(array, "hi")
        <Array [[1.1, 'hi', 2.2], [], ['hi', ...]] type='3 * var * union[float64, s...'>

    The list content now has a union type:

        >>> ak.fill_none(array, "hi").type.show()
        3 * var * union[
            float64,
            string
        ]

    The values could be floating-point numbers or strings.
    """
    # Dispatch
    yield array, value

    # Implementation
    return _impl(array, value, axis, highlevel, behavior)


def _impl(array, value, axis, highlevel, behavior):
    axis = regularize_axis(axis)
    arraylayout = ak.operations.to_layout(array, allow_record=True, allow_other=False)
    behavior = behavior_of(array, value, behavior=behavior)
    backend = backend_of(arraylayout, default=cpu)

    # Convert value type to appropriate layout
    if (
        isinstance(value, np.ndarray)
        and issubclass(value.dtype.type, (np.bool_, np.number))
        and len(value.shape) != 0
    ):
        valuelayout = ak.operations.to_layout(
            backend.nplike.asarray(value)[np.newaxis],
            allow_record=False,
            allow_other=False,
        )
    elif isinstance(value, (bool, numbers.Number, np.bool_, np.number)) or (
        isinstance(value, np.ndarray)
        and issubclass(value.dtype.type, (np.bool_, np.number))
    ):
        valuelayout = ak.operations.to_layout(
            backend.nplike.asarray(value), allow_record=False, allow_other=False
        )
    elif (
        is_sized_iterable(value)
        and not (isinstance(value, (str, bytes)))
        or isinstance(value, (ak.highlevel.Record, ak.record.Record))
    ):
        valuelayout = ak.operations.to_layout(
            value, allow_record=True, allow_other=False
        )
        if isinstance(valuelayout, ak.record.Record):
            valuelayout = valuelayout.array[valuelayout.at : valuelayout.at + 1]
        elif len(valuelayout) == 0:
            offsets = ak.index.Index64(
                backend.index_nplike.asarray([0, 0], dtype=np.int64)
            )
            valuelayout = ak.contents.ListOffsetArray(offsets, valuelayout)
        else:
            valuelayout = ak.contents.RegularArray(valuelayout, len(valuelayout), 1)
    else:
        valuelayout = ak.operations.to_layout(
            [value], allow_record=False, allow_other=False
        )

    if axis is None:

        def action(layout, continuation, **kwargs):
            if layout.is_option:
                return ak._do.fill_none(continuation(), valuelayout)

    else:

        def action(layout, depth, **kwargs):
            posaxis = maybe_posaxis(layout, axis, depth)
            if posaxis is not None and posaxis + 1 == depth:
                if layout.is_option:
                    return ak._do.fill_none(layout, valuelayout)
                elif layout.is_union or layout.is_record or layout.is_indexed:
                    return None
                else:
                    return layout

            elif layout.is_leaf:
                raise AxisError(
                    f"axis={axis} exceeds the depth of this array ({depth})"
                )

    out = ak._do.recursively_apply(arraylayout, action, behavior)
    return wrap_layout(out, behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import pytest  # noqa: F401

import awkward as ak


def test_array_with_length():
    array = ak.Array(
        [[100.0, 200.0, 22.0], [4.0, 5.0], [8.0, 9.0, 10.0, 11.0]], backend="typetracer"
    )
    counts = ak.Array([2, 1, 1, 1, 2, 2])
    ak.unflatten(array, counts)


def test_array_without_length():
    array = ak.Array(
        ak.to_layout(
            [[100.0, 200.0, 22.0], [4.0, 5.0], [8.0, 9.0, 10.0, 11.0]]
        ).to_typetracer(forget_length=True)
    )
    counts = ak.Array(
        ak.to_layout([2, 1, 1, 1, 2, 2]).to_typetracer(forget_length=True)
    )
    ak.unflatten(array, counts)


def test_scalar():
    array = ak.Array(
        [[100.0, 200.0, 22.0, 8.0], [4.0, 5.0, 5.0, 8.0], [8.0, 9.0, 10.0, 8.0]],
        backend="typetracer",
    )
    ak.unflatten(array, 2)


def test_unknown_scalar():
    array = ak.Array(
        ak.to_layout([[100, 200, 22], [4, 5, 5], [8, 9, 10]]).to_typetracer(
            forget_length=True
        )
    )
    ak.unflatten(array, array[0, 0])


def test_unknown_length():
    array = ak.Array(
        ak.to_layout(
            [[100.0, 200.0, 22.0], [4.0, 5.0, 5.0], [8.0, 9.0, 10.0]]
        ).to_typetracer(forget_length=True)
    )
    ak.unflatten(array, array.layout.length)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
__all__ = ("merge_union_of_records",)
import awkward as ak
from awkward._backends.numpy import NumpyBackend
from awkward._behavior import behavior_of
from awkward._dispatch import high_level_function
from awkward._errors import AxisError
from awkward._layout import maybe_posaxis, wrap_layout
from awkward._nplikes.numpylike import ArrayLike, NumpyMetadata
from awkward._regularize import regularize_axis

np = NumpyMetadata.instance()
cpu = NumpyBackend.instance()


@high_level_function
def merge_union_of_records(array, axis=-1, *, highlevel=True, behavior=None):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        axis (int): The dimension at which this operation is applied.
            The outermost dimension is `0`, followed by `1`, etc., and negative
            values count backward from the  innermost: `-1` is the innermost
            dimension, `-2` is the next level up, etc.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Simplifies unions of records, e.g.

        >>> array = ak.concatenate(([{"a": 1}], [{"b": 2}]))
        >>> array
        <Array [{a: 1}, {b: 2}] type='2 * union[{a: int64}, {b: int64}]'>

    into records of options, i.e.

        >>> ak.merge_union_of_records(array)
        <Array [{a: 1, b: None}, {a: None, ...}] type='2 * {a: ?int64, b: ?int64}'>

    Missing records are preserved in the result, e.g.

        >>> array = ak.concatenate(([{"a": 1}], [{"b": 2}, None]))
        >>> array
        <Array [{a: 1}, {b: 2}, None] type='3 * union[{a: int64}, ?{b: int64}]'>
        >>> ak.merge_union_of_records(array)
        <Array [{a: 1, b: None}, {...}, None] type='3 * ?{a: ?int64, b: ?int64}'>
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, axis, highlevel, behavior)


def _impl(array, axis, highlevel, behavior):
    axis = regularize_axis(axis)
    behavior = behavior_of(array, behavior=behavior)
    layout = ak.to_layout(array, allow_record=False)

    def invert_record_union(
        tags: ArrayLike, index: ArrayLike, contents, *, backend
    ) -> ak.contents.RecordArray:
        index_nplike = backend.index_nplike
        # First, create an ordered list containing the union of all fields
        seen_fields = set()
        all_fields = []
        for content in contents:
            # Find new fields
            for field in content.fields:
                if field not in seen_fields:
                    seen_fields.add(field)
                    all_fields.append(field)

        # Build unions for each field
        outer_field_contents = []
        for field in all_fields:
            field_tags = index_nplike.asarray(tags, copy=True)
            field_index = index_nplike.asarray(index, copy=True)

            # Build contents for union representing current field
            field_contents = [c.content(field) for c in contents if c.has_field(field)]

            # Find the best location for option type.
            # We will potentially have fewer contents in this per-field union
            # than the original outer union-of-records, because some recordarrays
            # may not have the given field.
            tag_for_missing = 0
            for i, content in enumerate(field_contents):
                if content.is_option:
                    tag_for_missing = i
                    break

            # If at least one recordarray doesn't have this field, we add
            # a special option
            if len(field_contents) < len(contents):
                # Make the tagged content an option, growing by one to ensure we
                # have a known `None` value to index into
                tagged_content = field_contents[tag_for_missing]
                indexedoption_index = backend.index_nplike.arange(
                    tagged_content.length + 1, dtype=np.int64
                )
                indexedoption_index[
                    index_nplike.shape_item_as_index(tagged_content.length)
                ] = -1
                field_contents[
                    tag_for_missing
                ] = ak.contents.IndexedOptionArray.simplified(
                    ak.index.Index64(indexedoption_index), tagged_content
                )

            # Index of None values in tagged content (content with extra None item at end)
            index_missing = index_nplike.shape_item_as_index(
                field_contents[tag_for_missing].length - 1
            )
            # Now build contents for union, by looping over outermost index
            # Overwrite tags to adjust for new contents length
            # and use the tagged content for any missing values
            k = 0
            for j, content in enumerate(contents):
                tag_is_j = field_tags == j

                if content.has_field(field):
                    # Rewrite tags to account for missing fields
                    field_tags[tag_is_j] = k
                    k += 1

                else:
                    # Rewrite tags to point to option content
                    field_tags[tag_is_j] = tag_for_missing
                    # Point each value to missing value
                    field_index[tag_is_j] = index_missing

            outer_field_contents.append(
                ak.contents.UnionArray.simplified(
                    ak.index.Index8(field_tags),
                    ak.index.Index64(field_index),
                    field_contents,
                )
            )
        return ak.contents.RecordArray(
            outer_field_contents, all_fields, backend=backend
        )

    def compact_option_index(index: ArrayLike, *, backend) -> ArrayLike:
        # Find dense (outer) index into non-null items.
        # This is in trivial order: the re-arranging is done by the union (below)
        is_none = index < 0
        num_none = backend.index_nplike.count_nonzero(is_none)
        dense_index = backend.index_nplike.empty(index.size, dtype=index.dtype)
        dense_index[is_none] = -1
        dense_index[~is_none] = backend.index_nplike.arange(
            index.size - num_none,
            dtype=index.dtype,
        )
        return dense_index

    def apply(layout, depth, backend, **kwargs):
        posaxis = maybe_posaxis(layout, axis, depth)
        if depth < posaxis + 1 and layout.is_leaf:
            raise AxisError(f"axis={axis} exceeds the depth of this array ({depth})")
        elif depth == posaxis + 1 and layout.is_union:
            if not all(
                x.is_record or x.is_indexed or x.is_option for x in layout.contents
            ):
                return

            # Any option types need to be re-written
            if any(x.is_option for x in layout.contents):
                # We'll rebuild the union to include only the non-null items.
                inner_union_index_parts = []
                next_contents = []
                next_tags_sparse = backend.index_nplike.asarray(layout.tags, copy=True)
                for tag, content in enumerate(layout.contents):
                    is_this_tag = backend.index_nplike.asarray(layout.tags) == tag

                    # Union arrays for this content
                    tag_index = backend.index_nplike.asarray(layout.index)[is_this_tag]

                    # For unmasked arrays, we can directly take the content
                    if isinstance(content, ak.contents.UnmaskedArray):
                        next_contents.append(content.content)
                        inner_union_index_parts.append(tag_index)
                    # Otherwise, we need to rebuild the index
                    elif content.is_option or content.is_indexed:
                        # Let's work with indexed option types for ease
                        if content.is_option:
                            content = content.to_IndexedOptionArray64()

                        # First, find the inner index that actually re-arranges the (non-null) items
                        content_index = backend.index_nplike.asarray(content.index)
                        merged_index = content_index[tag_index]
                        is_non_null = merged_index >= 0
                        inner_union_index_parts.append(merged_index[is_non_null])
                        # Mask out tags of items that are missing
                        next_tags_sparse[is_this_tag] = backend.index_nplike.where(
                            is_non_null, tag, -1
                        )

                        # The length of this index/option content is irrelevant; the union provides the length
                        next_contents.append(content.content)
                    # Non-indexed/option types are trivially included as-is
                    else:
                        next_contents.append(content)
                        inner_union_index_parts.append(tag_index)

                # We'll create an outermost indexed-option type, which re-instates the missing values.
                # This should have the same length as the original union, and its index should be "dense"
                # (contiguous, monotonic integers; or -1). Therefore, we can directly compute it from the "sparse"
                # tags index, which has the same length as the original union, and has only missing items set to -1.
                outer_option_dense_index = compact_option_index(
                    next_tags_sparse, backend=backend
                )

                # Ignore missing items for inner union, creating a dense array of tags
                next_tags = next_tags_sparse[next_tags_sparse >= 0]
                # Build dense index from parts for each tag
                next_index = backend.index_nplike.empty(next_tags.size, dtype=np.int64)
                for tag, content_index in enumerate(inner_union_index_parts):
                    next_index[next_tags == tag] = content_index

                # Return option around record of unions
                return ak.contents.IndexedOptionArray(
                    ak.index.Index64(outer_option_dense_index),
                    invert_record_union(
                        next_tags, next_index, next_contents, backend=backend
                    ),
                )

            # Any index types need to be re-written
            elif any(x.is_indexed for x in layout.contents):
                # We'll create an outermost indexed-option type, which re-instates the missing values
                current_index = backend.index_nplike.asarray(layout.index)
                next_index = backend.index_nplike.empty(
                    current_index.size, dtype=np.int64
                )

                # We'll rebuild the union to include only the non-null items.
                next_contents = []
                for tag, content in enumerate(layout.contents):
                    is_this_tag = backend.index_nplike.asarray(layout.tags) == tag

                    # Rewrite union index of indexed types
                    if content.is_indexed:
                        content_index = backend.index_nplike.asarray(content.index)
                        next_index[is_this_tag] = content_index[
                            current_index[is_this_tag]
                        ]
                        next_contents.append(content.content)

                    else:
                        next_index[is_this_tag] = current_index[is_this_tag]
                        next_contents.append(content)

                return invert_record_union(
                    backend.index_nplike.asarray(layout.tags),
                    next_index,
                    next_contents,
                    backend=backend,
                )

            else:
                return invert_record_union(
                    backend.index_nplike.asarray(layout.tags),
                    backend.index_nplike.asarray(layout.index),
                    layout.contents,
                    backend=backend,
                )

    out = ak._do.recursively_apply(layout, apply)
    return wrap_layout(out, highlevel=highlevel, behavior=behavior)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest

import awkward as ak


@pytest.mark.parametrize(
    "array",
    [
        ak.contents.NumpyArray(np.arange(4)),
        ak.contents.IndexedArray(
            ak.index.Index64(np.arange(4, dtype=np.int64)),
            ak.contents.NumpyArray(np.arange(4)),
        ),
        ak.contents.IndexedOptionArray(
            ak.index.Index64(np.arange(4, dtype=np.int64)),
            ak.contents.NumpyArray(np.arange(4)),
        ),
        ak.contents.ListOffsetArray(
            ak.index.Index64(np.arange(4, dtype=np.int64)),
            ak.contents.NumpyArray(np.arange(3)),
        ),
        ak.contents.ListArray(
            ak.index.Index64(np.arange(3, dtype=np.int64)),
            ak.index.Index64(np.arange(1, 4, dtype=np.int64)),
            ak.contents.NumpyArray(np.arange(3)),
        ),
        ak.contents.RegularArray(ak.contents.NumpyArray(np.arange(12)), size=3),
    ],
)
def test_highlevel_lowlevel(array):
    layout = ak.to_layout(array)
    assert isinstance(ak.type(layout), ak.types.ArrayType)
    # Check that the highlevel of ak.Array yields low level from form
    assert layout.form.type == ak.type(layout).content


def test_array():
    array = ak.Array(["this", {"x": ["is", 1, 2, None]}])
    assert ak.type(array) == array.type
    assert isinstance(array.type, ak.types.ArrayType)


def test_record():
    record = ak.Record({"y": ["this", {"x": ["is", 1, 2, None]}]})
    assert ak.type(record) == record.type
    assert isinstance(record.type, ak.types.ScalarType)


def test_none():
    assert ak.type(None) == ak.types.ScalarType(ak.types.UnknownType())


def test_unknown():
    with pytest.raises(TypeError):
        ak.type(object())


def test_bare_string():
    assert ak.type("hello") == ak.types.ArrayType(
        ak.types.NumpyType("uint8", parameters={"__array__": "char"}, typestr="char"), 5
    )
    assert ak.type(b"hello") == ak.types.ArrayType(
        ak.types.NumpyType("uint8", parameters={"__array__": "byte"}, typestr="byte"), 5
    )


def test_array_string():
    assert ak.type(["hello"]) == ak.types.ArrayType(
        ak.types.ListType(
            ak.types.NumpyType(
                "uint8", parameters={"__array__": "char"}, typestr="char"
            ),
            parameters={"__array__": "string"},
            typestr="string",
        ),
        1,
    )
    assert ak.type([b"hello"]) == ak.types.ArrayType(
        ak.types.ListType(
            ak.types.NumpyType(
                "uint8", parameters={"__array__": "byte"}, typestr="byte"
            ),
            parameters={"__array__": "bytestring"},
            typestr="bytes",
        ),
        1,
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np  # noqa: F401
import pytest  # noqa: F401

import awkward as ak


def test():
    one = ak.highlevel.Array([[{"x": 1}], [], [{"x": 2}]], with_name="One")
    two = ak.highlevel.Array([[{"x": 1.1}], [], [{"x": 2.2}]], with_name="Two")
    assert (
        str(
            ak.operations.with_name(
                ak.operations.concatenate([one, two], axis=1), "All"
            ).type
        )
        == "3 * var * All[x: float64]"
    )
    assert (
        str(
            ak.operations.with_name(
                ak.operations.concatenate([one[1:], two[1:]], axis=1),
                "All",
            ).type
        )
        == "2 * var * All[x: float64]"
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

raise ModuleNotFoundError(
    """The awkward._v2 submodule was provided for early access to awkward>=2, as it developed.

Now that version 2 has been released, awkward._v2 is no longer needed.

If you were an early adopter using

    import awkward._v2 as ak

you can replace it with

    import awkward as ak

or

    try:
        import awkward._v2 as ak   # provides v2 in 1.8.0rc1<=awkward<=1.10.1
    except ModuleNotFoundError:
        import awkward as ak       # provides v2 in awkward>=2

It is no longer possible to access v1 and v2 in the same process.

Arrays can be written and read as Parquet files or Arrow buffers to share data between processes."""
)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest

import awkward as ak

numba = pytest.importorskip("numba")

ak.numba.register_and_check()


def test_array():
    @numba.njit
    def f1(x):
        return np.array(x)

    assert isinstance(
        f1(ak.highlevel.Array([[[1], [2], [3]], [[4], [5], [6]]])), np.ndarray
    )
    assert f1(ak.highlevel.Array([[[1], [2], [3]], [[4], [5], [6]]])).tolist() == [
        [[1], [2], [3]],
        [[4], [5], [6]],
    ]
    assert f1(ak.highlevel.Array([[1, 2, 3], [4, 5, 6]])).tolist() == [
        [1, 2, 3],
        [4, 5, 6],
    ]
    assert f1(ak.highlevel.Array([1, 2, 3, 4, 5, 6])).tolist() == [1, 2, 3, 4, 5, 6]

    with pytest.raises(ValueError):
        f1(ak.highlevel.Array([[1, 2, 3, 4], [5, 6]]))


def test_asarray():
    @numba.njit
    def f1(x):
        return np.asarray(x[-1][1:])

    akarray = ak.highlevel.Array(
        [[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6, 7.7, 8.8, 9.9]]
    )
    nparray = f1(akarray)

    assert nparray.tolist() == [7.7, 8.8, 9.9]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import pytest

import awkward as ak


def test():
    array = ak.Array([[1, 2, 3], [4, 5, 6]], backend="typetracer")

    with pytest.raises(IndexError):
        array[0, 0, 0]

    with pytest.raises(IndexError):
        array[0, :, 0]

    with pytest.raises(IndexError):
        array[0, :, ..., 0]

    with pytest.raises(IndexError):
        array[0, [0], [0]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
from __future__ import annotations

from awkward._typing import Self, TypeAlias

ShapeItem: TypeAlias = "int | type[unknown_length]"


class _UnknownLength:
    _name: str

    @classmethod
    def _new(cls, name: str) -> Self:
        self = super().__new__(cls)
        self._name = name
        return self

    def __new__(cls, *args, **kwargs):
        raise TypeError(
            "internal_error: the _UnknownLength class should never be directly instantiated"
        )

    def __add__(self, other) -> Self | NotImplemented:
        if isinstance(other, int) or other is self:
            return self
        else:
            return NotImplemented

    __radd__ = __add__
    __iadd__ = __add__

    def __sub__(self, other) -> Self | NotImplemented:
        if isinstance(other, int) or other is self:
            return self
        else:
            return NotImplemented

    __rsub__ = __sub__
    __isub__ = __sub__

    def __mul__(self, other) -> Self | NotImplemented:
        if isinstance(other, int) or other is self:
            return self
        else:
            return NotImplemented

    __rmul__ = __mul__
    __imul__ = __mul__

    def __floordiv__(self, other) -> Self | NotImplemented:
        if isinstance(other, int) or other is self:
            return self
        else:
            return NotImplemented

    __rfloordiv__ = __floordiv__
    __ifloordiv__ = __floordiv__

    def __divmod__(self, other) -> tuple[Self, Self]:
        return self, self

    __rdivmod__ = __divmod__

    def __str__(self) -> str:
        return "##"

    def __repr__(self):
        return f"{__name__}.{self._name}"

    def __eq__(self, other) -> bool:
        if other is self:
            return True
        else:
            raise TypeError("cannot compare unknown lengths against known values")

    def __gt__(self, other):
        raise TypeError("cannot order unknown lengths")

    def __index__(self):
        raise TypeError("cannot interpret unknown lengths as concrete index values")

    def __int__(self):
        raise TypeError("cannot interpret unknown lengths as concrete values")

    __bool__ = __int__
    __float__ = __int__

    __ge__ = __gt__
    __le__ = __gt__
    __lt__ = __gt__


unknown_length = _UnknownLength._new("unknown_length")
